import re
import sys
import json
import unittest

from pedal import Submission
from pedal.core.final_feedback import FinalFeedback
from pedal.core.submission import create_ipynb_submission_parser
from pedal.resolvers.core import make_resolver
from pedal.source import verify, next_section as original_next_section
from pedal.core.feedback import Feedback
from pedal.core.environment import Environment
from pedal.core.report import MAIN_REPORT
from pedal.sandbox import run, get_sandbox, set_input, start_trace, get_output
from pedal.source.sections import FeedbackSourceSection
from pedal.tifa import tifa_analysis
from pedal.resolvers.simple import resolve as simple_resolve, by_priority
from pedal.core.formatting import Formatter, HtmlFormatter
from pedal.resolvers.silent import resolve as silent_resolve


def get_source_code():
    from pedal.environments.jupyter import EXTRACT_STUDENT_CODE, EXECUTE_CODE
    from IPython.display import display, HTML, Javascript
    from inspect import currentframe, getframeinfo
    import inspect
    from pprint import pprint
    frames = inspect.stack()
    for frame in frames:
        if frame.function == '<module>':
            break
    split_line = str(frame.lineno)
    get_source_code_js = Javascript("""
    console.log("HEY WAIT WHY");
    var output_area = this;
    // find my cell element
    var cell_element = output_area.element.parents('.cell');
    // which cell is it?
    var cells = Jupyter.notebook.get_cells();
    var cell_idx = Jupyter.notebook.get_cell_elements().index(cell_element);
    // TODO: Walk backwards to get last "Autograded Answer" cell
    // get the cell object
    let cell = null;
    for (let i= cell_idx; i > 0; i -= 1) {
        cell = Jupyter.notebook.get_cell(i);
        if (cell.metadata.nbgrader && cell.metadata.nbgrader.solution) {
            break;
        }
    }
    let instructor_cell = Jupyter.notebook.get_cell(cell_idx);
    let instructor_text = instructor_cell.get_text();
    instructor_text = instructor_text.split("\\n")
                                     .slice(""" + split_line + """)
                                     .join("\\n");
    console.log('>>>', instructor_text);
    let source_code = "";
    if (cell !== null) {{
        source_code = cell.get_text();
    }}

    var kernel = Jupyter.notebook.kernel;
    var get_source_code = "student_code="+JSON.stringify(source_code);
    get_source_code += `
from pedal.environments.nbgrader import NBGraderEnvironment
NBGraderEnvironment(main_code=student_code)
if True:
`+instructor_text;
    console.log(get_source_code);
    if (kernel !== null) {
        console.log("DOUBLE CHECK");
        console.log(kernel.execute(get_source_code, {
            iopub: {
                output: (data) => {
                    //console.log(">>>", data, output_area, instructor_cell);
                    if (data.msg_type !== "error" || data.content.ename !== "PedalNotebookException") {
                        output_area.handle_output(data);
                    } else {
                        console.error("Pedal Exception:", data);
                    }
                    /*
                    console.log('+++', data);
                    console.log(output_area);
                    console.log(cell_element);
                    console.log(instructor_cell);
                    data.output_type = data.msg_type;
                    instructor_cell.outputs.push(data);
                    if (data.msg_type == "error") {
                        // If this was an error, show the traceback properly.
                        output_area.append_error(data.content);
                        console.error(data);
                    } else if (data.msg_type === "display_data") {
                        // If it was valid data, we show it as HTML.
                        console.log(data);
                        output_area.append_display_data(data.content);
                        instructor_cell.outputs.push(data.content);
                        //output_area.element.html(output_area.element.html()+"\\n"+
                        //                         data.content.text.replace(/\\n/g, "<br>"));
                    } else {
                        // I'm not sure what it is - better dump it on the console.
                        console.log(data);
                    }
                */}
            }
        }));
    }
    console.log("FINISHED SETTING UP JS");
    """)

    # class UglyException(Exception):
    #    pass

    # def handle(*args, **kwargs):
    #    import ctypes
    #    frame = inspect.stack()[2].frame
    #    frame.f_locals['outflag'] = True
    #    ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_bool(True))

    # from IPython import get_ipython
    # get_ipython().set_custom_exc((UglyException,), handle)

    display(get_source_code_js)
    # import time
    # time.sleep(4)
    # raise UglyException()
    # print("FINISHED SET UP")
    return ""
    # return os.environ['STUDENT_SOURCE_CODE'] #student_code


from io import StringIO


class PedalNotebookException(SystemExit):
    """Exit Exception for IPython.

    Exception temporarily redirects stderr to buffer.
    """

    def __init__(self):
        # print("exiting")  # optionally print some message to stdout, too
        # ... or do other stuff before exit
        sys.stderr = StringIO()

    def __del__(self):
        sys.stderr.close()
        sys.stderr = sys.__stderr__  # restore from backup


def ipy_exit():
    raise PedalNotebookException


def in_notebook():
    """
    Returns ``True`` if the module is running in IPython kernel,
    ``False`` if in IPython shell or other Python shell.
    """
    from IPython import get_ipython
    return bool(get_ipython())
    # return 'ipykernel' in sys.modules #or 'IPython' in sys.modules


class NBGraderEnvironment(Environment):
    """
    Configures the NBGrader programming environment.
    """

    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='on_run.py', skip_tifa=True, skip_run=True,
                 inputs=None, set_correct=True, set_success=None,
                 report=MAIN_REPORT, trace=True):
        super().__init__(files=files, main_file=main_file, main_code=main_code,
                         user=user, assignment=assignment, course=course,
                         execution=execution, instructor_file=instructor_file,
                         report=report)
        # TODO: If WebCAT doesn't want HTML output, then we should switch the default formatter
        self.report.add_hook('pedal.report.add_feedback', self.force_immediate_errors)
        self.skip_run = skip_run
        self.skip_tifa = skip_tifa
        self.trace = trace
        report.set_formatter(Formatter(report))
        verify(report=self.report)
        if not skip_tifa:
            tifa_analysis(report=self.report)
        if inputs:
            set_input(inputs)
        if skip_run:
            student = get_sandbox(report=report)
        else:
            if trace:
                start_trace()
            student = run(report=report)
        self.fields = {
            'student': student,
            'resolve': silent_resolve,
        }

    def force_immediate_errors(self, feedback, report=None):
        if report is None:
            report = self.report
        if self.report.get_current_group() is not None:
            return False

        final = simple_resolve()
        tc = unittest.TestCase()
        with tc.subTest():
            tc.fail(final.message)

        # if in_notebook():
        #    from IPython.display import HTML, display
        #    display(HTML(f"""<strong>{final.title}</strong><br><div>{final.message}</div>"""))
        #    ipy_exit()
        # else:
        #    print(final.title, file=sys.stderr)
        #    print(final.message, file=sys.stderr)
        #    sys.exit(1)

    def load_main(self, path):
        """
        Allowed to return either a string value (the contents of the file)
        or the exception that was raised.
        """
        try:
            return super().load_main(path)
        except OSError as file_error:
            return file_error


def setup_environment(cells=None, **kwargs):
    """

    Args:
        cells: The subset of cells to get. Defaults to ALL cells (None, or '**'), including the non-solution cells.
            To limit to just solution cells, use '*'. You can also pass a single cell ID or a list of cell IDs
            to grab a specific subset of the notebook.
        ipynb_file:
        main_code:
        main_file:
        **kwargs:

    Returns:

    """
    if in_notebook():
        get_source_code()
        return False
    else:
        # Logic for parsing the `cells` parameter
        all_cells, all_solutions = False, False
        chosen_cells = []
        if cells is None or cells == '**':
            all_cells = True
        else:
            if isinstance(cells, str):
                chosen_cells.append(cells)
            else:
                chosen_cells.extend(cells)
            for cell in chosen_cells[:]:
                if cell == '*':
                    all_solutions = True
            # TODO: Support "<=" and ">=" to get adjacent cells?

        def metadata_filter(cell, other_cells) -> bool:
            if all_cells:
                return True
            nbgrader = cell.metadata.get('nbgrader', {})
            if all_solutions and nbgrader.get('solution', False):
                return True
            return nbgrader.get('grade_id') in chosen_cells

        Submission.PARSERS['ipynb'] = create_ipynb_submission_parser(metadata_filter)
        environment = NBGraderEnvironment(**kwargs)
        return environment

# # Old approach: inspecting the stack to get filename - yikes!
# import inspect
# student_filename = inspect.stack()[-1].filename
# with open(student_filename) as student_file:
#    main_code = student_file.read()



"""
Simple enumeration of the Modes available for Pedal's command line.
"""
import json
import os
import sys
import traceback
from contextlib import redirect_stdout
from io import StringIO
import unittest
from textwrap import indent
from unittest.mock import patch
from pprint import pprint
import warnings
import argparse

from pedal.command_line.report import StatReport
from pedal.command_line.verify import generate_report_out, ReportVerifier
from pedal.core.report import MAIN_REPORT
from pedal.core.submission import Submission
from pedal.utilities.files import normalize_path, find_possible_filenames
from pedal.utilities.progsnap import SqlProgSnap2
from pedal.utilities.text import chomp
from pedal.resolvers.export import PedalJSONEncoder, clean_json

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(values):
        """ Trivial helper function to replace TQDM's behavior """
        yield from values


# Disable all warnings from student files
# TODO: Move this strictly into ICS and Submission execution
warnings.filterwarnings("ignore")


def is_progsnap(path):
    """ Determines if this is a ProgSnap file or folder. """
    if os.path.isfile(path):
        name, extension = os.path.splitext(path)
        return extension in ('.zip', '.db')
    else:
        # Does it have all the necessary components?
        return os.path.isfile(os.path.join(path, 'MainTable.csv'))


def get_python_files(paths):
    result = {}
    for path in paths:
        name, ext = os.path.splitext(path)
        if ext != ".py":
            continue
        with open(path, 'r') as submission_file:
            contents = submission_file.read()
        result[path] = contents
    return result


class BundleResult:
    def __init__(self, data, output, error, resolution):
        self.data = data
        self.output = output
        self.error = error
        self.resolution = resolution

    def to_json(self):
        resolution = self.resolution.copy() if self.resolution else {}
        #if 'considered' in resolution:
        #    for c in resolution['considered']:
        #        if 'fields' in c:
        #            del c['fields']
        return dict(
            output=self.output,
            error=self.error,
            **resolution
        )

class Bundle:
    def __init__(self, config, script, submission):
        self.config = config
        self.script = script
        self.submission = submission
        self.environment = None
        self.result = None

    def to_json(self):
        return dict(
            script=self.script,
            submission=self.submission.to_json(),
            environment=self.environment,
            result=self.result.to_json()
        )

    def run_ics_bundle(self, resolver='resolve', skip_tifa=False, skip_run=False):
        """
        Runs the given instructor control script on the given submission, with the
        accompany contextualizations.
        """
        ics_args = [self.submission, self.environment]
        captured_output = StringIO()
        global_data = {}
        error = None
        resolution = None
        if self.environment:
            env = __import__('pedal.environments.' + self.environment,
                             fromlist=[''])
            global_data.update(env.setup_environment(self.submission,
                                                     skip_tifa=skip_tifa,
                                                     skip_run=skip_run,
                                                     threaded=self.config.threaded).fields)
        else:
            MAIN_REPORT.contextualize(self.submission)
        with redirect_stdout(captured_output):
            with patch.object(sys, 'argv', ics_args):
                try:
                    grader_exec = compile(self.script,
                                          self.submission.instructor_file, 'exec')
                    exec(grader_exec, global_data)
                    #print(repr(self.script), file=x)
                    #print(list(global_data.keys()), file=x)
                    if 'MAIN_REPORT' in global_data:
                        if not global_data['MAIN_REPORT'].resolves:
                            if resolver in global_data:
                                resolution = global_data[resolver]()
                            # TODO: Need more elegance/configurability here
                            elif self.config.resolver == 'resolve':
                                exec("from pedal.resolvers import print_resolve", global_data)
                                resolution = global_data["print_resolve"]()
                        else:
                            resolution = global_data['MAIN_REPORT'].resolves[-1]
                except Exception as e:
                    error = e
        actual_output = captured_output.getvalue()
        self.result = BundleResult(global_data, actual_output, error, resolution)


class AbstractPipeline:
    """ Generic pipeline for handling all the phases of executing instructor
    control scripts on submissions, and reformating the output. """

    def __init__(self, config):
        if isinstance(config, dict):
            # TODO: Include default argument automatically
            config = argparse.Namespace(**config)
        self.config = config
        self.submissions = []
        self.result = None

    def execute(self):
        self.load_submissions()
        self.setup_execution()
        self.run_control_scripts()
        return self.process_output()

    def load_file_submissions(self, scripts):
        # Get instructor control scripts
        all_scripts = []
        for script in scripts:
            script_file_name, script_file_extension = os.path.splitext(script)
            # Single Python file
            if script_file_extension in ('.py',):
                with open(script, 'r') as scripts_file:
                    scripts_contents = scripts_file.read()
                all_scripts.append((script, scripts_contents))
        given_submissions = self.config.submissions
        # If submission is a directory, use it as a directory adjacent to each ics
        if os.path.isdir(given_submissions):
            for script, scripts_contents in all_scripts:
                directory_pattern = given_submissions
                submission_dir = normalize_path(directory_pattern, script)
                submission_files = [
                    os.path.join(submission_dir, sub)
                    for sub in os.listdir(submission_dir)
                ]
                subs = get_python_files(submission_files)
                for main_file, main_code in subs.items():
                    new_submission = Submission(
                        main_file=main_file, main_code=main_code,
                        instructor_file=script
                    )
                    self.submissions.append(Bundle(self.config, scripts_contents, new_submission))
        # Otherwise, if the submission is a single file:
        # Maybe it's a Progsnap DB file?
        elif given_submissions.endswith('.db'):
            for script, scripts_contents in all_scripts:
                self.load_progsnap(given_submissions, instructor_code=scripts_contents)
        # Otherwise, must just be a single python file.
        else:
            main_file = given_submissions
            load_error, possible_load_error = None, None
            alternatives = [given_submissions]
            # if alternative filenames given, we'll queue them up
            if self.config.alternate_filenames:
                alternatives.extend(find_possible_filenames(self.config.alternate_filenames))
            # Run through all possible filenames
            for possible in alternatives:
                try:
                    with open(possible, 'r') as single_submission_file:
                        main_code = single_submission_file.read()
                        main_file = possible
                    break
                except OSError as e:
                    # Only capture the first possible load error
                    if possible_load_error is None:
                        possible_load_error = e
            else:
                # Okay, file does not exist. Load error gets triggered.
                main_code = None
                load_error = possible_load_error
            for script, scripts_contents in all_scripts:
                new_submission = Submission(
                    main_file=main_file, main_code=main_code,
                    instructor_file=script, load_error=load_error
                )
                self.submissions.append(Bundle(self.config, scripts_contents, new_submission))
            return load_error

    progsnap_events_map = {
        'run': 'Run.Program',
        'compile': 'Compile',
        'edit': 'File.Edit',
        'last': 'File.Edit'
    }

    def load_progsnap(self, path, instructor_code=None):
        script_file_name, script_file_extension = os.path.splitext(path)
        if script_file_extension in ('.db',):
            with SqlProgSnap2(path, cache=self.config.cache) as progsnap:
                if self.config.progsnap_profile:
                    progsnap.set_profile(self.config.progsnap_profile)
                link_filters = {}
                include_scripts = self.config.include_scripts
                if include_scripts:
                    link_filters['Assignment'] = {}
                    if include_scripts.startswith('name='):
                        link_filters['Assignment']['X-Name'] = include_scripts[5:]
                    elif include_scripts.startswith('id='):
                        link_filters['Assignment']['AssignmentId'] = include_scripts[3:]
                    elif include_scripts.startswith('url='):
                        link_filters['Assignment']['X-URL'] = include_scripts[4:]
                    else:
                        link_filters['Assignment']['X-URL'] = include_scripts
                event_type = self.progsnap_events_map[self.config.progsnap_events]
                events = progsnap.get_events(event_filter={'EventType': event_type},
                                             link_filters=link_filters, limit=self.config.limit)
            if self.config.progsnap_events == 'last':
                runs_by_user_assign = {}
                for event in sorted(events, key=lambda e: e['event_id']):
                    key = (event[progsnap.PROFILES[progsnap.profile]['link_primary']['user']], event['assignment_name'])
                    runs_by_user_assign[key] = event
                events = list(runs_by_user_assign.values())
            for event in events:
                if instructor_code is None:
                    instructor_code_for_this_run = event['on_run']
                else:
                    instructor_code_for_this_run = instructor_code
                link_selections = progsnap._merge('link_selections', {})
                new_submission = Submission(
                    main_file='answer.py',
                    main_code=event['submission_code'] if isinstance(event['submission_code'], str) else
                        event['submission_code'].decode('utf-8'),
                    instructor_file='instructor.py',
                    #files={'cisc106.py': 'from bakery import *'},
                    execution=dict(client_timestamp=event['client_timestamp'],
                                   event_id=event['event_id']),
                    #user=dict(email=event['student_email'],
                    #          first=event['student_first'],
                    #          last=event['student_last']),
                    user={key: event[key] for key in link_selections['Subject'].values()}
                    if 'Subject' in link_selections else {'id': event['subject_id']},
                    assignment=dict(name=event['assignment_name'],
                                    url=event['assignment_url']),
                )
                self.submissions.append(Bundle(self.config, instructor_code_for_this_run, new_submission))
            # raise ValueError("TODO: ProgSnap DB files not yet supported")
            # Progsnap Zip
        elif script_file_extension in ('.zip',):
            raise ValueError("TODO: Zip files not yet supported")

    def load_submissions(self):
        given_script = self.config.instructor
        if self.config.ics_direct:
            # TODO: Allow non-progsnap ics_direct
            self.load_progsnap(self.config.submissions, instructor_code=given_script)
        elif is_progsnap(given_script):
            self.load_progsnap(given_script)
        elif os.path.isfile(given_script):
            self.load_file_submissions([given_script])
        elif os.path.isdir(given_script):
            python_files = os.listdir(given_script)
            self.load_file_submissions(python_files)
        else:
            potential_filenames = list(find_possible_filenames(given_script))
            for filename in potential_filenames:
                load_error = self.load_file_submissions([filename])
                if load_error is None:
                    return
            from pedal.source.feedbacks import source_file_not_found
            source_file_not_found(potential_filenames[0], False)

    def setup_execution(self):
        for bundle in self.submissions:
            bundle.environment = self.config.environment

    def run_control_scripts(self):
        for bundle in self.submissions:
            bundle.run_ics_bundle(resolver=self.config.resolver,
                                  skip_tifa=self.config.skip_tifa,
                                  skip_run=self.config.skip_run)

    def process_output(self):
        for bundle in self.submissions:
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file,
                  bool(bundle.result.error))


class FeedbackPipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            #print(bundle.submission.instructor_file,
            #      bundle.submission.main_file)
            if bundle.result.error:
                print(bundle.result.error)
            elif bundle.result.resolution:
                # report = bundle.result.data['MAIN_REPORT']
                #print(bundle.result.resolution.title)
                #print(bundle.result.resolution.message)
                print(bundle.result.output)
            else:
                print("Error: No feedback determined.")


class RunPipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file)
            if bundle.result.error:
                print(bundle.result.error)
            elif bundle.result.resolution:
                report = bundle.result.data['MAIN_REPORT']
                generate_report_out(None, self.config.environment, report)
            else:
                print("No feedback determined.")


class StatsPipeline(AbstractPipeline):
    def run_control_scripts(self):
        for bundle in tqdm(self.submissions):
            bundle.run_ics_bundle(resolver='stats_resolve', skip_tifa=self.config.skip_tifa,
                                  skip_run=self.config.skip_run)

    def process_output(self):
        total = 0
        errors = 0
        final = []
        for bundle in self.submissions:
            if bundle.result.error:
                print(bundle.result.error)
                errors += 1
            else:
                final.append(bundle.to_json())
            total += 1
        final = clean_json(final)
        if self.config.output is not None:
            #print(final)
            print("Total Processed:", total)
            print("Errors:", errors)
            pedal_json_encoder = PedalJSONEncoder(indent=2, skipkeys=True)
            if self.config.output == 'stdout':
                print(pedal_json_encoder.encode(final))
            else:
                with open(self.config.output, 'w') as output_file:
                    print(pedal_json_encoder.encode(final), file=output_file)
        return StatReport(final)



class VerifyPipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            bundle.run_ics_bundle(resolver=self.config.resolver, skip_tifa=self.config.skip_tifa,
                                  skip_run=self.config.skip_run)
            if bundle.result.error:
                raise bundle.result.error
            # Now get the expected output
            base = os.path.splitext(bundle.submission.main_file)[0]
            test_name = os.path.basename(base)
            if os.path.exists(base + ".out"):
                verifier = ReportVerifier(base + ".out", self.config.environment)
                self.add_case(test_name, bundle.result, verifier)
            else:
                if self.config.create_output:
                    report = bundle.result.data['MAIN_REPORT']
                    generate_report_out(base + ".out", self.config.environment, report)
                    self.skip_case(test_name)
                else:
                    self.error_case(test_name, bundle.result)
        self.run_cases()

    def __init__(self, config):
        super().__init__(config)

        class TestReferenceSolutions(unittest.TestCase):
            __module__ = 'VerifyTestCase'
            __qualname__ = config.instructor_name
            maxDiff = None

            def __str__(self):
                return f"{config.instructor}, using {self._testMethodName[5:]}.py"
                # return f"{self.id().__name__} (Test{config.instructor})"

            def __repr__(self):
                return str(self)

        self.TestReferenceSolutions = TestReferenceSolutions
        self.tests = []

    def add_case(self, name, bundle_result, verifier):
        """ Add this ``name`` as a new test. """
        if bundle_result.output.strip():
            print(bundle_result.output)
        error = bundle_result.error if bundle_result.error else None
        if error:
            print(error)
        if 'MAIN_REPORT' not in bundle_result.data:
            # TODO: Better error message here
            raise Exception("No MAIN_REPORT found; did you import Pedal?")
        actual = bundle_result.data['MAIN_REPORT'].result.to_json()
        error = bundle_result.error
        expected_fields = verifier.get_final()

        def new_test(self):
            if error:
                raise error
            entire_actual, entire_expected = "", ""
            differing_fields = []
            for field, value in expected_fields:
                self.assertIn(field, actual, msg="Field was not in feedback.")
                actual_value = chomp(str(actual[field]).strip())
                expected_value = chomp(str(value))
                entire_actual += field + ": " + actual_value + "\n"
                entire_expected += field + ": " + expected_value + "\n"
                if expected_value != actual_value:
                    differing_fields.append(field)
            differing_fields = ', '.join(f"'{field}'" for field in differing_fields)
            self.assertEqual(entire_expected, entire_actual,
                             msg=f"Wrong value for {differing_fields} in {name}.")

        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def skip_case(self, name):
        def new_test(self):
            self.skipTest("Output did not exist; created.")

        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def error_case(self, name, bundle_result):
        def new_test(self):
            self.fail(
                "Expected output file was not found next to the instructor file. Perhaps you meant to use --create_output?")

        setattr(self.TestReferenceSolutions, "test_" + name, new_test)
        self.tests.append("test_" + name)

    def run_cases(self):
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(self.TestReferenceSolutions)
        runner = unittest.TextTestRunner()
        runner.run(suite)


class GradePipeline(AbstractPipeline):
    def process_output(self):
        if self.config.output == 'stdout':
            self.print_bundles(sys.stdout)
        else:
            with open(self.config.output, 'w') as output_file:
                self.print_bundles(output_file)

    def print_bundles(self, target):
        #print(len(self.submissions))
        for bundle in self.submissions:
            print(bundle.result.output, file=target)
            if bundle.result.error:
                raise bundle.result.error
            # This info is not sent to the output target, just to stdout
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file,
                  bundle.submission.user.get('student_email') if bundle.submission.user else 'Unknown User',
                  bundle.submission.assignment.get('name') if bundle.submission.assignment else 'Unknown Assignment',
                  1 if bundle.result.resolution.correct else
                  bundle.result.resolution.score,
                  sep=", ", end="")


class SandboxPipeline(AbstractPipeline):
    """ Run the given script in a sandbox. """

    ICS = """from pedal import *
verify()
run()"""

    def load_submissions(self):
        # Use first argument as student submissions
        given_script = self.config.instructor
        # ... either a single file
        if os.path.isfile(given_script):
            scripts = [given_script]
        # ... or a directory of files
        else:
            scripts = os.listdir(given_script)
        # Then create submission to run
        for script in scripts:
            script_file_name, script_file_extension = os.path.splitext(script)
            if script_file_extension in ('.py',):
                load_error = None
                try:
                    with open(script, 'r') as scripts_file:
                        scripts_contents = scripts_file.read()
                except Exception as e:
                    load_error = e
                new_submission = Submission(
                    main_file="answer.py", main_code=scripts_contents,
                    instructor_file=script, load_error=load_error
                )
                self.submissions.append(Bundle(self.config, self.ICS, new_submission))

    def process_output(self):
        # Print output
        # Print runtime exceptions, if any
        for bundle in self.submissions:
            if bundle.result.error:
                traceback.print_tb(bundle.result.error.__traceback__)
                print(bundle.result.error)
                continue
            sandbox = bundle.result.data['MAIN_REPORT']['sandbox']['sandbox']
            output = [sandbox.raw_output]
            if sandbox.feedback:
                output.append(sandbox.feedback.message)
            print("\n".join(output))


class DebugPipeline(AbstractPipeline):
    def process_output(self):
        for bundle in self.submissions:
            print(bundle.submission.instructor_file,
                  bundle.submission.main_file)
            print("****** Student Code:")
            print(indent(bundle.submission.main_code, "    "))
            print("****** Results")
            if bundle.result.error:
                print(bundle.result.error)
            else:
                print("Output:")
                print(indent(bundle.result.output, " " * 4))
                report = bundle.result.data['MAIN_REPORT']
                print("Feedback:")
                for feedback in report.feedback:
                    print("\t", feedback, repr(feedback.fields))
                print("Final Feedback")
                pprint(report.result.to_json())


class MODES:
    """
    The possible modes that Pedal can be run in.
    """
    # Runs the instructor control script and outputs all feedback
    RUN = 'run'
    # Get just the feedback for submissions
    FEEDBACK = 'feedback'
    # Get just the score/correctness
    GRADE = 'grade'
    # Check against expected output
    VERIFY = 'verify'
    # Output all the feedback objects
    STATS = 'stats'
    # Run the file as student code using the sandbox
    SANDBOX = 'sandbox'
    # Output as much information as possible
    DEBUG = 'debug'

    PIPELINES = {
        RUN: RunPipeline,
        FEEDBACK: FeedbackPipeline,
        GRADE: GradePipeline,
        STATS: StatsPipeline,
        SANDBOX: SandboxPipeline,
        VERIFY: VerifyPipeline,
        DEBUG: DebugPipeline
    }



""" Tests for some AST operations stuffed into CAIT """
import unittest

from tests.execution_helper import Execution, SUCCESS_TEXT
from pedal.cait import *
from pedal.cait.find_node import (is_top_level, find_function_calls, function_is_called, find_prior_initializations,
                                  find_operation)


class TestUtilities(unittest.TestCase):
    def test_is_top_level(self):
        with Execution('print("Test")\ndef a(x):\n  print(x+1)\na(1)') as e:
            ast = parse_program()
            defs = ast.find_all('FunctionDef')
            self.assertEqual(len(defs), 1)
            self.assertTrue(is_top_level(defs[0]))
            self.assertEqual(len(defs[0].body), 1)
            self.assertFalse(is_top_level(defs[0].body[0]))
            calls = ast.find_all('Call')
            self.assertEqual(len(calls), 3)
            self.assertTrue(is_top_level(calls[0]))
            self.assertFalse(is_top_level(calls[1]))
        self.assertEqual(e.final.message, SUCCESS_TEXT)

    def test_find_function_calls(self):
        with Execution('def a(x):\n  print(x,1)\na(1)\nprint("T")') as e:
            prints = find_function_calls('print')
            self.assertEqual(len(prints), 2)
            self.assertEqual(len(prints[0].args), 2)
            self.assertEqual(len(prints[1].args), 1)
        self.assertEqual(e.final.message, SUCCESS_TEXT)
        with Execution('a=[]\na.append(a)\na.pop()\n') as e:
            pops = find_function_calls('pop')
            self.assertEqual(len(pops), 1)
            self.assertEqual(len(pops[0].args), 0)
        self.assertEqual(e.final.message, SUCCESS_TEXT)

    def test_function_is_called(self):
        with Execution('a=[]\na.append(a)\na.pop()\n') as e:
            self.assertTrue(function_is_called('pop'))
            self.assertFalse(function_is_called('print'))
        self.assertEqual(e.final.message, SUCCESS_TEXT)

    def test_find_prior_initializations(self):
        with Execution('a=0\na\na=5\na') as e:
            ast = parse_program()
            self.assertEqual(len(ast.body), 4)
            self.assertEqual(ast.body[3].ast_name, "Expr")
            self.assertEqual(ast.body[3].value.ast_name, "Name")
            priors = find_prior_initializations(ast.body[3].value)
            self.assertEqual(len(priors), 2)


    def test_find_operation(self):
        with Execution('1+1') as e:
            ast = parse_program()
            self.assertNotEqual(find_operation("+", ast), False)
        with Execution('1>1') as e:
            ast = parse_program()
            self.assertNotEqual(find_operation(">", ast), False)
        with Execution('True and True') as e:
            ast = parse_program()
            self.assertNotEqual(find_operation("and", ast), False)
        with Execution('not True') as e:
            ast = parse_program()
            self.assertNotEqual(find_operation("not", ast), False)
        with Execution('not (1 + 1) and 1 < 1 <= 10') as e:
            ast = parse_program()
            self.assertFalse(find_operation(">", ast))
        with Execution('1 in [1,2,3]') as e:
            ast = parse_program()
            self.assertNotEqual(find_operation("in", ast), False)



from pedal.core.feedback import Feedback
from pedal.core.feedback_category import FeedbackCategory, FeedbackKind
from pedal.questions.constants import TOOL_NAME


class show_question(Feedback):
    """

    Args:
        instructions:
        report:
    """
    category = FeedbackCategory.INSTRUCTIONS
    kind = FeedbackKind.INSTRUCTIONAL
    justification = 'Question was asked but not answered.'
    title = 'Show Question'
    priority = 'instructions'
    tool = TOOL_NAME
    valence = 0

    def __init__(self, instructions, **kwargs):
        super().__init__(message=instructions, **kwargs)



import unittest
import os
import sys
from textwrap import dedent

pedal_library = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, pedal_library)

here = "" if os.path.basename(os.getcwd()) == "tests" else "tests/"

from pedal.questions.loader import check_question
from tests.execution_helper import Execution, ExecutionTestCase

ADD5_QUESTION = {'functions': [{
    'name': 'add5',
    'parameters': ["int"],
    'cases': [{
        "arguments": [10],
        "returns": 15
    }, {
        "arguments": [20],
        "returns": 25
    }]
}]}

ADD5_QUESTION_SIGNATURE = {'functions': [{
    'name': 'add5',
    'signature': "int -> int",
    'cases': [{
        "arguments": [10],
        "returns": 15
    }, {
        "arguments": [20],
        "returns": 25
    }]
}]}

LIST_SIGNATURE = {'functions': [{
    'name': 'summate',
    'signature': "[int] -> int",
    'cases': [{
        "arguments": [[40, 20, 30, 10]],
        "returns": 100
    }]
}]}

@unittest.skip
class TestQuestionsLoader(ExecutionTestCase):

    def test_loader_missing_function(self):
        with Execution('def undefined():\n pass\nundefined()', run_tifa=False) as e:
            check_question(ADD5_QUESTION)
        self.assertFeedback(e, "Instructor Feedback\nNo function named `add5` was found.")

    def test_loader_fewer_parameters(self):
        with Execution('def add5():\n pass\nadd5()') as e:
            check_question(ADD5_QUESTION)
        self.assertFeedback(e, "Instructor Feedback\nThe function named `add5` has fewer parameters (0) than expected (1).")

    def test_loader_more_parameters(self):
        with Execution('def add5(a, b):\n a\n b\nadd5(1, 2)') as e:
            check_question(ADD5_QUESTION)
        self.assertFeedback(e, "Instructor Feedback\nThe function named `add5` has more parameters (2) than expected (1).")

    def test_loader_failing_tests(self):
        with Execution('def add5(a):\n return 7+a-a\nadd5(1)') as e:
            check_question(ADD5_QUESTION)
        self.assertFeedback(e, """Instructor Feedback
I ran your function <code>add5</code> on my own test cases. It failed 2/2 of my tests.
<table class='pedal-test-cases table table-sm table-bordered table-hover'>
    <tr class='table-active'>
        <th></th>
        <th>Arguments</th>
        <th>Expected</th>
        <th>Returned</th>
    </tr>
        <tr>
        <td>&#10060;</td>
        <td><code>10</code></td>
        <td><code>15</code></td>
        <td><code>7</code></td>
    </tr>
    <tr>
        <td>&#10060;</td>
        <td><code>20</code></td>
        <td><code>25</code></td>
        <td><code>7</code></td>
    </tr>
</table>""")

    def test_loader_error_tests(self):
        with Execution('def add5(a):\n a+""\nadd5("1")') as e:
            check_question(ADD5_QUESTION)
        self.assertFeedback(e, """Instructor Feedback
I ran your function <code>add5</code> on my own test cases. It failed 2/2 of my tests.
<table class='pedal-test-cases table table-sm table-bordered table-hover'>
    <tr class='table-active'>
        <th></th>
        <th>Arguments</th>
        <th>Expected</th>
        <th>Returned</th>
    </tr>
        <tr>
        <td>&#10060;</td>
        <td><code>10</code></td>
        <td colspan='2'>Error: <code>unsupported operand type(s) for +: 'int' and 'str'</code></td>
    </tr>
    <tr>
        <td>&#10060;</td>
        <td><code>20</code></td>
        <td colspan='2'>Error: <code>unsupported operand type(s) for +: 'int' and 'str'</code></td>
    </tr>
</table>""")

    def test_loader_success(self):
        with Execution('def add5(a):\n return a+5\nadd5(3)') as e:
            check_question(ADD5_QUESTION)
        self.assertFeedback(e, "Complete\nGreat work!")

    # TODO: Finish this test once the functionality is ready!
    @unittest.skip
    def test_loading_signature_no_parameter_type(self):
        with Execution('def add5(a):\n return a+5\nadd5(3)') as e:
            check_question(ADD5_QUESTION_SIGNATURE)
        self.assertFeedback(e, "Instructor Feedback\nError in definition of function `add5` signature. Expected `int -> int`, instead found `Any -> Any`.")

    # TODO: Finish this test once the functionality is ready!
    @unittest.skip
    def test_loading_signature_no_return_type(self):
        with Execution('def add5(a: int):\n return a+5\nadd5(3)') as e:
            check_question(ADD5_QUESTION_SIGNATURE)
        self.assertFeedback(e, "Instructor Feedback\nError in definition of function `add5` signature. Expected `int -> int`, instead found `int -> Any`.")

    # TODO: Finish this test once the functionality is ready!
    @unittest.skip
    def test_loading_signature_correct_type(self):
        with Execution('def add5(a: int) -> int:\n return a+5\nadd5(3)') as e:
            check_question(ADD5_QUESTION_SIGNATURE)
        self.assertFeedback(e, "Complete\nGreat work!")

    # TODO: Finish this test once the functionality is ready!
    @unittest.skip
    def test_loading_list_question(self):
        with Execution('def summate(vals: [int]) -> int:\n return sum(vals)\nsummate([1,2,3])') as e:
            check_question(LIST_SIGNATURE)
        self.assertFeedback(e, "Complete\nGreat work!")

if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Functions for letting students organize their code into sections.
"""

import re

from pedal.core.feedback import FeedbackGroup
from pedal.core.report import MAIN_REPORT
from pedal.source.constants import TOOL_NAME
from pedal.source.feedbacks import not_enough_sections, incorrect_number_of_sections
from pedal.source.substitutions import Substitution

DEFAULT_SECTION_PATTERN = r'^(##### Part .+)$'


class FeedbackSourceSection(FeedbackGroup):
    tool = TOOL_NAME
    category = FeedbackGroup.CATEGORIES.SYSTEM
    message = "Feedback separated into groups"

    def __init__(self, section_number, **kwargs):
        self.section_number = section_number
        super().__init__(section_number, **kwargs)


def separate_into_sections(pattern=DEFAULT_SECTION_PATTERN, independent=True, report=MAIN_REPORT):
    """
    Chunks the current submissions' main code into separate sections.
    Args:
        pattern:
        report:

    Returns:

    """
    if not report[TOOL_NAME]['success']:
        pass
    if report[TOOL_NAME]['sections']:
        # TODO: System constraint violated: separating into sections multiple times
        pass
    report[TOOL_NAME]['independent'] = independent
    report[TOOL_NAME]['section'] = 0
    report[TOOL_NAME]['section_group'] = FeedbackSourceSection(0)
    report.start_group(report[TOOL_NAME]['section_group'])
    report.submission.clear_line_offsets()
    report[TOOL_NAME]['section_pattern'] = pattern
    report[TOOL_NAME]['sections'] = re.split(pattern, report.submission.main_code, flags=re.MULTILINE)

    backup = Substitution(report.submission.main_code, report.submission.main_file)
    report[TOOL_NAME]['substitutions'].append(backup)
    report.submission.replace_main(report[TOOL_NAME]['sections'][0])

    report.add_hook('pedal.resolvers.resolve', stop_any_sections)

    #print(report[TOOL_NAME]['sections'])


def _calculate_section_number(section_index):
    return int((section_index+1)/2)


def stop_sections(report=MAIN_REPORT):
    """

    Args:
        report:
    """
    # TODO: If no substitutions, then throw error that we haven't separated into sections
    old_submission = report[TOOL_NAME]['substitutions'].pop()
    report.stop_group(report[TOOL_NAME]['section_group'])
    report.submission.replace_main(old_submission.code, old_submission.filename)
    report[TOOL_NAME]['section_group'] = None

def stop_any_sections(report=MAIN_REPORT):
    if report[TOOL_NAME]['substitutions']:
        stop_sections(report)


def next_section(name="", report=MAIN_REPORT):
    """

    TODO: Incoporate name to allow instructors to name sections (defaults to numbers)

    Args:
        name:
        report:
    """
    report.execute_hooks(TOOL_NAME, 'next_section.before')
    source = report[TOOL_NAME]
    old_submission = report[TOOL_NAME]['substitutions'][-1]
    report.stop_group(report[TOOL_NAME]['section_group'])
    report.submission.replace_main(old_submission.code, old_submission.filename)
    # Advance to next section
    source['section'] += 2
    section_index = source['section']
    section_number = _calculate_section_number(section_index)
    report[TOOL_NAME]['section_group'] = FeedbackSourceSection(section_number)
    sections = source['sections']
    found = _calculate_section_number(len(source['sections']))
    if section_number <= found:
        if source['independent']:
            new_code = ''.join(sections[section_index])
            old_code = ''.join(sections[:section_index])
            report.submission.set_line_offset(len(old_code.split("\n"))-1)
        else:
            new_code = ''.join(sections[:section_index + 1])
        report.submission.replace_main(new_code)
        report[TOOL_NAME]['success'] = None
        report.start_group(report[TOOL_NAME]['section_group'])
    else:
        not_enough_sections(section_number, found)
    report.execute_hooks(TOOL_NAME, 'next_section.after')


def check_section_exists(section_number, report=MAIN_REPORT):
    """
    Checks that the right number of sections exist. The prologue before the
    first section is 0, while subsequent ones are 1, 2, 3, etc. 
    So if you have 3 sections in your code plus the prologue,
    you should pass in 3 and not 4 to verify that all of them exist.
    """
    if report[TOOL_NAME]['success'] is False:
        return False
    found = int((len(report[TOOL_NAME]['sections']) - 1) / 2)
    #print(section_number, found, len(report[TOOL_NAME]['sections']))
    if section_number > found:
        incorrect_number_of_sections(section_number, found, parent=report[TOOL_NAME]['section'], report=report)



from textwrap import dedent
import unittest
from pprint import pprint
import os
import sys
import json

from pedal.core.submission import Submission

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pedal.core.commands import MAIN_REPORT, clear_report, contextualize_report
from pedal.sandbox import Sandbox, run
import pedal.sandbox.commands as commands
from pedal.source import set_source
from pedal.utilities.system import IS_AT_LEAST_PYTHON_311, IS_AT_LEAST_PYTHON_310

here = "" if os.path.basename(os.getcwd()) == "tests" else "tests/"

class TestCode(unittest.TestCase):
    def setUp(self):
        """

        """
        pass

    def test_normal_run(self):
        student = Sandbox()
        student.run('a=0\nprint(a)', filename='student.py')
        self.assertIn('a', student.data)
        self.assertEqual(student.data['a'], 0)
        self.assertEqual(len(student.output), 1)
        self.assertIn('0', student.output[0])

    def test_input(self):
        student = Sandbox()
        student.run('b = input("Give me something:")\nprint(b)',
                    inputs=['Hello World!'])
        self.assertIn('b', student.data)
        self.assertEqual(student.data['b'], 'Hello World!')

    def test_oo(self):
        # Load the "bank.py" code
        student_code = dedent('''
            class Bank:
                def __init__(self, balance):
                    self.balance = balance
                def save(self, amount):
                    self.balance += amount
                    return self.balance > 0
                def take(self, amount):
                    self.balance -= amount
                    return self.balance > 0''')
        student = Sandbox()
        student.run(student_code, filename='bank.py')
        # Check that we created the class
        self.assertIn('Bank', student.data)
        # Now let's try making an instance
        student.call('Bank', 50, target='bank')
        self.assertIsInstance(student.data['bank'], student.data['Bank'])
        # Can we save money?
        student.call('bank.save', 32)
        self.assertTrue(student.result)
        # What about extracting money?
        student.data['bank'].balance += 100
        student.call('bank.take', 100)
        self.assertTrue(student.result)

    def test_improved_exceptions(self):
        TEST_FILENAME = here+'_sandbox_test_student.py'
        with open(TEST_FILENAME) as student_file:
            student_code = student_file.read()
        student = Sandbox()
        student.run(student_code, filename=TEST_FILENAME)
        self.assertIsNotNone(student.exception)
        self.assertIsInstance(student.exception, TypeError)
        self.assertEqual(1, student.feedback.location.line)
        self.assertEqual(dedent(
            """I ran the file _sandbox_test_student.py.

A TypeError occurred:

    Unsupported operand type(s) for +: 'int' and 'str'

The traceback was:
Line 1 of file _sandbox_test_student.py
    1+'0'
""" + ('    ^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + """
Type errors occur when you use an operator or function on the wrong type of value. For example, using `+` to add to a list (instead of `.append`), or dividing a string by a number.

Suggestion: To fix a type error, you should trace through your code. Make sure each expression has the type you expect it to have.
""".format(filename=TEST_FILENAME)).strip(), student.feedback.message)

    def test_error_context(self):
        student_code = dedent('''
                    def get_input():
                        return int(input("Gimme the number"))
                ''')
        set_source(student_code, "student.py")
        student = Sandbox()
        student.run()
        result = student.call("get_input", inputs="Banana")
        print("--", [c.inputs for c in student._context])
        self.assertEqual(3, student.feedback.location.line)
        self.assertEqual(dedent(
            """I ran the code:
    get_input()

And I entered as input:
    Banana 
    
A ValueError occurred:

    Invalid literal for int() with base 10: 'Banana'

The traceback was:
Line 3 of file student.py in get_input
        return int(input("Gimme the number"))
""" + ('               ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + """
A ValueError occurs when you pass the wrong type of value to a function. For example, you try to convert a string without numbers to an integer (like `int('Five')`).

Suggestion: Read the error message to see which function had the issue. Check what type the function expects. Then trace your code to make sure you pass in that type.""").strip(), student.feedback.message)

    def test_call(self):
        student_code = "def average(a,b):\n return (a+b)/2"
        student = Sandbox()
        student.run(student_code, filename='student.py')
        student.call('average', 10, 12)

    def test_commands_api(self):
        clear_report()
        student_code = 'word = input("Give me a word")\nprint(word+"!")'
        set_source(student_code)
        self.assertFalse(commands.get_output())
        commands.queue_input("Hello")
        self.assertIsInstance(commands.run(), Sandbox)
        self.assertEqual(["Give me a word", "Hello!"],
                         commands.get_output())
        commands.queue_input("World", "Again")
        self.assertIsInstance(commands.run(), Sandbox)
        self.assertEqual(commands.get_output(),
                         ["Give me a word", "Hello!",
                          "Give me a word", "World!"])
        self.assertIsInstance(commands.run(), Sandbox)
        self.assertEqual(commands.get_output(),
                         ["Give me a word", "Hello!",
                          "Give me a word", "World!",
                          "Give me a word", "Again!"])
        commands.reset_output()
        commands.queue_input("Dogs", "Are", "Great")
        self.assertIsInstance(commands.run(), Sandbox)
        self.assertIsInstance(commands.run(), Sandbox)
        self.assertIsInstance(commands.run(), Sandbox)
        self.assertEqual(commands.get_output(),
                         ["Give me a word", "Dogs!",
                          "Give me a word", "Are!",
                          "Give me a word", "Great!"])
        
        
        commands.reset_output()
        commands.queue_input(json.dumps("Virginia,Trend"))
        self.assertIsInstance(commands.run(), Sandbox)
        self.assertEqual(commands.get_output(),
                         ["Give me a word", '"Virginia,Trend"!'])

    def test_commands_exceptions(self):
        student_code = '1 + "Banana"'
        set_source(student_code)
        commands.run()
        self.assertIsNotNone(commands.get_exception())


    def test_good_unique_calls(self):
        student_code = dedent("""
        def x(n):
            if n > 0:
                return x(n-1)
            return 0
        x(5)
        x(4)
        """)
        set_source(student_code)
        commands.start_trace()
        commands.run()
        self.assertEqual(commands.count_unique_calls('x'), 6)

    def test_tracing_returns(self):
        student_code = dedent("""
        def x(n):
            if n > 0:
                return x(n-1)*2
            return 1
        x(5)
        x(4)""")
        set_source(student_code)
        commands.start_trace()
        student = commands.run()
        self.assertEqual(len(student.trace.calls), 1)
        self.assertEqual(len(student.trace.calls['x']), 11)
        self.assertEqual(len(student.trace.returns), 1)
        self.assertEqual(len(student.trace.returns['x']), 11)
        self.assertEqual(student.trace.calls['x'][0], {'n': 5})
        self.assertEqual(student.trace.returns['x'][0], [32])

    def test_get_by_types(self):
        student_code = dedent('''
            my_int = 0
            my_other_int = 5
            my_str = "Hello there!"
            response_str = "General Kenobi!"
            a_list = [1,2,3]
            another_list = [4,5,6]
        ''')
        student = Sandbox()
        student.run(student_code, filename='student.py')
        # ints
        ints = student.get_names_by_type(int)
        self.assertEqual(len(ints), 2)
        self.assertIn('my_int', ints)
        self.assertIn('my_other_int', ints)
        # lists
        lists = student.get_names_by_type(list)
        self.assertEqual(len(lists), 2)
        self.assertIn('a_list', lists)
        self.assertIn('another_list', lists)
        # strs
        strs = student.get_values_by_type(str)
        self.assertEqual(len(strs), 2)
        self.assertIn('Hello there!', strs)
        self.assertIn('General Kenobi!', strs)

    def test_matplotlib(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.plot([1,2,3])
            plt.title("My line plot")
            plt.show()
        ''')
        contextualize_report(student_code)
        student = run(student_code, filename='student.py')
        self.assertIn('plotting', dir(student.modules))
        plt = student.modules.plotting
        self.assertEqual(len(plt.plots), 1)
        self.assertIsNone(student.exception)

    def test_matplotlib_commands(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.plot([1,2,3])
            plt.title("My line plot")
            plt.show()
            plt.hist([1,2,3])
            plt.show()
        ''')
        contextualize_report(student_code)
        student = commands.run()
        print(student)
        plt2 = student.modules.plotting
        self.assertEqual(len(plt2.plots), 2)

    def test_blocked_module_import(self):
        student_code = dedent('''
            import os
            os
        ''')
        set_source(student_code)
        exception = commands.run()
        self.assertIsNotNone(exception)

    def test_coverage(self):
        test_filename = here+'_sandbox_test_coverage.py'
        with open(test_filename) as student_file:
            student_code = student_file.read()
        contextualize_report(student_code)
        student = Sandbox()
        student.tracer_style = 'native'
        student.run(student_code, filename=test_filename)
        self.assertIsNone(student.exception)
        #self.assertEqual(round(student.trace.pc_covered), 85)

        if IS_AT_LEAST_PYTHON_310:
            self.assertEqual([1, 2, 3, 4, 5, 6, 9, 12, 16, 14, 15, 17], student.trace.lines)
        else:
            self.assertEqual([1, 2, 3, 4, 6, 9, 12, 16, 14, 15, 17], student.trace.lines)


    def test_coverage_native(self):
        test_filename = here+'_sandbox_test_coverage.py'
        with open(test_filename) as student_file:
            student_code = student_file.read()
        contextualize_report(student_code)
        student = Sandbox()
        student.tracer_style = 'native'
        student.run(student_code, filename=test_filename)
        self.assertIsNone(student.exception)
        if IS_AT_LEAST_PYTHON_310:
            self.assertEqual(student.trace.lines, [1, 2, 3, 4, 5, 6, 9, 12, 16, 14, 15, 17])
        else:
            self.assertEqual(student.trace.lines, [1, 2, 3, 4, 6, 9, 12, 16, 14, 15, 17])
        self.assertEqual(student.trace.calls, {'z': [{'args': (0,), 'banana': 2, 'kwargs': {'apple': 8}, 'p': 3}]})

    def test_calls(self):
        student_code = dedent('''
            def x(n):
                if n > 0:
                    return x(n-1)
                return 0
            x(5)
        ''')
        student = Sandbox()
        student.tracer_style = 'calls'
        student.run(student_code, filename='student.py')
        self.assertEqual(len(student.trace.calls['x']), 6)


    def test_runtime_error_inputs(self):
        student_code = dedent('''
                    def x():
                        value = input("Gimme a number")
                        return 7 % value
                    x
                ''')
        student = Sandbox()
        set_source(student_code)
        student.run()
        result = student.call("x", inputs=["0"])

        self.assertEqual(("""I ran the code:
    x()

And I entered as input:
    0 

A TypeError occurred:

    Unsupported operand type(s) for %: 'int' and 'str'

The traceback was:
Line 4 of file answer.py in x
        return 7 % value
""" + ('               ^^^^^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") +"""
Type errors occur when you use an operator or function on the wrong type of value. For example, using `+` to add to a list (instead of `.append`), or dividing a string by a number.

Suggestion: To fix a type error, you should trace through your code. Make sure each expression has the type you expect it to have.
""").strip(), student.feedback.message)

    def test_unittest(self):
        student_code = dedent('''
            x = 0
        ''')
        student = Sandbox()
        student.run(student_code)
        student.call('x')
        self.assertIsNotNone(student.exception)

        student_code = dedent('''
        class Fruit:
            def __init__(self, name, weight=0):
                self.name = name
                self.weight = weight
        def do_math(a, b):
            return a + b - 5
        def weigh_fruits(fruits):
            return sum(fruit.weight for fruit in fruits)    
        ''')
        student.report.contextualize(Submission(main_code=student_code))
        student = Sandbox()
        student.run()
        result = student.call('do_math', 15, 20)
        self.assertEqual(result, 30)
        self.assertEqual(['do_math(15, 20)'],
                         [context.code for context in student.get_context()])
        banana = student.call('Fruit', "Banana")
        self.assertIsInstance(banana, student.data['Fruit'])
        self.assertEqual(["Fruit('Banana')"],
                         [context.code for context in student.get_context()])
        student.start_grouping_context()
        student.run()
        orange = student.call("Fruit", "Orange", 30, target="orange")
        self.assertIsInstance(orange, student.data['Fruit'])
        student.call("Fruit", "Pineapple", 60, target="pineapple")
        student.run("fruits = [orange, pineapple]")
        total_weight = student.call('weigh_fruits', args_locals=["fruits"])
        self.assertEqual(total_weight, 90)
        self.assertEqual([student_code,
                          "orange = Fruit('Orange', 30)",
                          "pineapple = Fruit('Pineapple', 60)",
                          "fruits = [orange, pineapple]",
                          "weigh_fruits(fruits)"],
                         [context.code for context in student.get_context()])
        # print(student.call('weigh_fruits', student.var['fruits']))
        # from pprint import pprint
        # pprint(student.call_contexts)

    def test_multiline_statements(self):
        student_code = dedent('''
            class X:
                def __init__(self, a, b):
                    self.a = a
                    self.b = b
                def update(self):
                    self.a += 10
        ''')
        # Need to be able to run several commands, any of which could fail
        #       needs to send out a nice stack trace in that case
        # We often want to assert the result of a call OR attribute OR other
        # data (type of attribute)
        # x = X(5, 10)
        # x.a == 5
        # x.b == 10
        # x.update()
        # x.a == 15
    
    def test_sandboxing_open(self):
        student_code = dedent('''
            print(open("test_sandbox.py").read())
        ''')
        set_source(student_code)
        student = Sandbox()
        student.run(student_code, filename='student.py')
        self.assertEqual(str(student.exception), "The filename you passed to 'open' is restricted.")
        # Then can we turn it back on?
        student.allow_function('open')
        student.run(student_code, filename='student.py')
        self.assertIsNone(student.exception)
        # And off again
        student.block_function('open')
        student.run(student_code, filename='student.py')
        self.assertEqual(str(student.exception), "You are not allowed to call 'open'.")

    def test_block_requests_module(self):
        student_code = dedent('''
            import os
            print(os.listdir())
        ''')
        set_source(student_code)
        student = Sandbox()
        student.block_module('os')
        student.run(student_code, filename='student.py')
        self.assertEqual(str(student.exception), "You cannot import `os` from student code.")
        # Then can we turn it back on?
        student.allow_module('os')
        student.run(student_code, filename='student.py')
        self.assertIsNone(student.exception)
        # And off again
        student.reset_default_overrides()
        student.block_module('os')
        student.run(student_code, filename='student.py')
        self.assertEqual(str(student.exception), "You cannot import `os` from student code.")

    def test_sandboxing_pedal(self):
        student_code = dedent('''
            from pedal.report import MAIN_REPORT
            print(MAIN_REPORT)
        ''')
        set_source(student_code)
        student = Sandbox()
        student.run(student_code, filename='student.py')
        self.assertEqual(str(student.exception), "You cannot import pedal!")
    
    def test_sandboxing_sys_modules(self):
        clear_report()
        student_code = dedent('''
            import sys
            # Might try to bypass us
            del sys.modules['pedal']
            from pedal.report import MAIN_REPORT
            print(MAIN_REPORT)
        ''')
        set_source(student_code)
        student = Sandbox()
        student.run(student_code, filename='student.py')
        self.assertEqual(str(student.exception), "You cannot import pedal!")
    
    def test_sandboxing_devious_open1(self):
        student_code = dedent('''
            __builtins__['globals']()['open']("test_sandbox.py")
            # Fine to import anything else
            
        ''')
        set_source(student_code)
        student = Sandbox()
        student.run(student_code, filename='student.py')
        self.assertEqual(str(student.exception), "You are not allowed to call 'globals'.")

    def test_mock_turtle(self):
        student_code = dedent('''
            import turtle
            turtle.forward(100)
            turtle.mainloop()
        ''')
        set_source(student_code)
        student = Sandbox()
        student.run(student_code, filename='student.py')
        self.assertIsNone(student.exception)
        self.assertEqual(student.modules.turtles.calls[0][0], 'forward')
        self.assertEqual(student.modules.turtles.calls[0][1][0], 100)

    def test_weird_module_behavior(self):
        student_code = dedent('''
                    import pprint
                    pprint.pprint(5)
                ''')
        set_source(student_code)
        student = Sandbox()
        student.run(student_code, filename='student.py')
        self.assertIsNone(student.exception)

    def test_range_requires_integer(self):
        contextualize_report("x = 5")
        commands.run()
        x = commands.evaluate("x")
        range(x)
        self.assertIsNone(commands.get_exception())

    def test_int_requires_integer(self):
        contextualize_report("x = '5'")
        commands.run()
        x = int(commands.evaluate("x"))
        self.assertIsNone(commands.get_exception())

    def test_block_exit(self):
        contextualize_report("exit()")
        commands.run()
        self.assertIsNotNone(commands.get_exception())

    @unittest.skip("TODO: Something weird with recursion")
    def test_bad_recursion(self):
        contextualize_report("""
def to_pig_latin(str):
    first = str[0]
    str = str[1:]
    str = str + first + "ay"
    to_pig_latin("hello")""")
        commands.run()
        commands.call('to_pig_latin', 'test', threaded=True)
        self.assertNotIsInstance(commands.get_exception(), RecursionError)

    def test_duplicate_parameters(self):
        contextualize_report("def x(y,y): pass")
        commands.run()
        self.assertIsNotNone(commands.get_exception())

    def test_imported_input_is_mocked(self):
        contextualize_report(Submission({
            "answer.py": "from another import call_input\nprint(call_input())",
            "another.py": "def call_input():\n return input()"
        }))
        #commands.mock_module("another", {"call_input": lambda: input()}, "another")
        commands.queue_input("XYZ")
        commands.run()
        self.assertIsNone(commands.get_exception())
        self.assertEqual(["", "XYZ"], commands.get_output())

    # TODO: test `import builtins` strategy to access original builtins

if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Unit tests from the old curriculum-sneks library that has now been ported into Pedal core.
"""
from textwrap import dedent
import unittest
from pedal import *
from pedal.assertions.static import *
from pedal.utilities.system import IS_AT_LEAST_PYTHON_39
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE


class TestFiles(ExecutionTestCase):

    def test_files_not_closed(self):
        with Execution('open("not opened.txt")') as e:
            self.assertTrue(files_not_handled_correctly(1))
        self.assertFeedback(e, "Unclosed Files\nYou have not closed all the files you "
                               "were supposed to.")

    def test_files_closed_as_function(self):
        with Execution('open("not opened.txt")\nclose()') as e:
            self.assertTrue(files_not_handled_correctly(1))
        self.assertFeedback(e, "Close Is a Method\nYou have attempted to call "
                               "`close` as a function, but it is "
                               "actually a method of the file object.")

    def test_files_open_without_filename(self):
        with Execution('open()') as e:
            self.assertTrue(files_not_handled_correctly(1))
        self.assertEqual(e.final.message, "You have called the `open` "
                                          "function without any arguments. It needs a "
                                          "filename.")

    def test_files_open_as_method(self):
        with Execution('"filename.txt".open()') as e:
            self.assertTrue(files_not_handled_correctly(1))
        self.assertEqual(e.final.message, "You have attempted to call "
                                          "`open` as a method, but it is actually a "
                                          "built-in function.")

    def test_files_not_all_files_opened(self):
        with Execution('a = open("A.txt")\na.close()') as e:
            self.assertTrue(files_not_handled_correctly(2))
        self.assertEqual(e.final.message, "You have not opened all the files you "
                                          "were supposed to.")

    def test_files_too_many_opens(self):
        with Execution('a = open("A.txt")\nb = open("B.txt")'
                       '\na.close()\nb.close()') as e:
            self.assertTrue(files_not_handled_correctly(1))
        self.assertEqual(e.final.message, "You have opened more files than you "
                                          "were supposed to.")

    def test_files_too_many_closes(self):
        with Execution('a = open("A.txt")\n\na.close()\na.close()') as e:
            self.assertTrue(files_not_handled_correctly(1))
        self.assertEqual(e.final.message, "You have closed more files than you "
                                          "were supposed to.")

    def test_files_too_many_with_opens(self):
        with Execution('with open("A.txt") as out:\n  b = open("B.txt")'
                       '\n  b.close()') as e:
            self.assertTrue(files_not_handled_correctly(1))
        self.assertEqual(e.final.message, "You have opened more files than you "
                                          "were supposed to.")

    def test_files_missing_filename(self):
        with Execution('with open("A.txt") as out:\n  print(out.read())') as e:
            self.assertTrue(files_not_handled_correctly("X.txt"))
            suppress(Feedback.CATEGORIES.RUNTIME)
        self.assertEqual(e.final.message, "You must use the literal value "
                                          "'X.txt'.")

    def test_files_with_open_success(self):
        with Execution('with open("A.txt") as out:\n  print(out.read())') as e:
            self.assertFalse(files_not_handled_correctly("A.txt"))
            suppress(Feedback.CATEGORIES.RUNTIME)
        self.assertEqual(e.final.message, "Great work!")

        with Execution('a = open("filename.txt")\na.close()') as e:
            self.assertFalse(files_not_handled_correctly(1))


class TestFiles(ExecutionTestCase):
    def test_only_printing_variables(self):
        with Execution('a,b=0,1\nprint(a,b)') as e:
            self.assertFalse(only_printing_variables())
        with Execution('print(0,"True", True)') as e:
            self.assertTrue(only_printing_variables())
        with Execution('print(True)') as e:
            self.assertTrue(only_printing_variables())


    def test_prevent_advanced_iteration(self):
        with Execution('while False:\n  pass') as e:
            prevent_advanced_iteration()
        self.assertEqual(e.final.message, "You used a while loop on line 1. You may not use that.")
        with Execution('sum([1,2,3])') as e:
            prevent_advanced_iteration()
        self.assertEqual(e.final.message, "You used the function sum on line 1. You may not use that function.")

    @unittest.skip
    def test_ensure_assignment(self):
        with Execution('a=0') as e:
            self.assertNotEqual(ensure_assignment("a", "Num"), False)
        with Execution('a=""') as e:
            self.assertNotEqual(ensure_assignment("a", "Str"), False)
        with Execution('a=True') as e:
            self.assertNotEqual(ensure_assignment("a", "Bool"), False)


from pedal.sandbox.mocked import MockModuleExposing, MethodExposer
from pedal.sandbox.mocked import MockModule
from pedal.utilities.system import IS_PYTHON_36


def friendly_urls(url: str) -> str:
    if url.strip("/") == "index":
        return "/"
    if not url.startswith('/'):
        url = '/' + url
    return url


class MockDrafter(MockModuleExposing):
    """
    Mock Drafter library that can be used to capture data from the students' program execution.
    """
    expose = MethodExposer()
    UNKNOWN_FUNCTIONS = []

    def __init__(self):
        super().__init__()
        self.original_routes = []
        self.routes = {}
        self._unknown_calls = []
        self.server_started = False

    def __repr__(self):
        return f"MockDrafter()"

    def __str__(self):
        return f"<MockDrafter()>"

    def add_route(self, url, func):
        self.original_routes.append((url, func))
        url = friendly_urls(url)
        self.routes[url] = func

    @expose
    def start_server(self):
        self.server_started = True

    @expose
    def show_debug_information(self):
        pass

    @expose
    def hide_debug_information(self):
        pass

    @expose
    def route(self, url: str = None):
        if callable(url):
            local_url = url.__name__
            self.add_route(local_url, url)
            return url

        def make_route(func):
            local_url = url
            if url is None:
                local_url = func.__name__
            self.add_route(local_url, func)
            return func

        return make_route

    @staticmethod
    def _load_drafter_component(name):

        def call_drafter_component(*args, **kwargs):
            import drafter
            return getattr(drafter, name)(*args, **kwargs)

        return call_drafter_component

    COMPONENTS = ['BulletedList', 'Button', 'CheckBox',
                  'Header', 'HorizontalRule', 'Image', 'LineBreak', 'Link', 'LinkContent',
                  'NumberedList', 'Page', 'PageContent', 'SelectBox',
                  'SubmitButton', 'Table', 'Text', 'TextArea', 'TextBox']
    for component in COMPONENTS:
        expose.add_with_name(_load_drafter_component(component), component)







"""
route, start_server
'show_debug_information', 
'hide_debug_information',
'route', 'start_server',
         
         'bold', 'change_background_color', 'change_border',
         'change_color', 'change_height', 'change_margin', 'change_padding', 'change_text_align',
         'change_text_decoration',
         'change_text_font', 'change_text_size', 'change_text_transform', 'change_width',
         'float_left',
         'float_right',  
         'italic', 
         'large_font', 'monospace',
         'small_font', 'strikethrough', 'underline',
         'update_attr', 'update_style']
"""


from pedal.sandbox.library.matplotlib import MockPlt
from pedal.sandbox.library.designer import MockDesigner
from pedal.sandbox.library.turtles import MockTurtle
from pedal.sandbox.library.microbit import MockMicrobit



import os
import sys
import io
from contextlib import redirect_stdout
import unittest

pedal_library = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, pedal_library)

here = "" if os.path.basename(os.getcwd()) == "tests" else "tests/"

from tests.execution_helper import ExecutionTestCase
from pedal.environments.nbgrader import setup_environment
from pedal.source import next_section, verify_section, check_section_exists, set_source_file, separate_into_sections
from pedal.core.commands import compliment, clear_report, contextualize_report, log
from pedal.assertions import assert_equal, prevent_ast

try:
    from IPython import get_ipython
    skip = False
except ModuleNotFoundError:
    skip = True


@unittest.skipIf(skip, "IPython not installed")
class TestNBGrader(ExecutionTestCase):

    def test_parse_file(self):
        with io.StringIO() as f, redirect_stdout(f):
            nbgrader = setup_environment("*", main_file=here+'datafiles/jupyter_notebook_example.ipynb')
            print(nbgrader.report.submission.main_code)
            self.assertEqual('''def squares(n):
    """Compute the squares of numbers from 1 to n, such that the 
    ith element of the returned list equals i^2.
    
    """
    ### BEGIN SOLUTION
    if n < 1:
        raise ValueError("n must be greater than or equal to 1")
    return [i ** 2 for i in range(1, n + 1)]
    ### END SOLUTION
''', f.getvalue())

    def test_parse_whole_file(self):
        with io.StringIO() as f, redirect_stdout(f):
            nbgrader = setup_environment("**", main_file=here + 'datafiles/jupyter_notebook_example.ipynb')
            print(nbgrader.report.submission.main_code)
            self.assertEqual('''"""For this problem set, we'll be using the Jupyter notebook:

![](jupyter.png)"""
"""---
## Part A (2 points)

Write a function that returns a list of numbers, such that $x_i=i^2$, for $1\leq i \leq n$. Make sure it handles the case where $n<1$ by raising a `ValueError`."""
def squares(n):
    """Compute the squares of numbers from 1 to n, such that the 
    ith element of the returned list equals i^2.
    
    """
    ### BEGIN SOLUTION
    if n < 1:
        raise ValueError("n must be greater than or equal to 1")
    return [i ** 2 for i in range(1, n + 1)]
    ### END SOLUTION
"""Your function should print `[1, 4, 9, 16, 25, 36, 49, 64, 81, 100]` for $n=10$. Check that it does:"""
squares(10)
"""Check that squares returns the correct output for several inputs"""
assert squares(1) == [1]
assert squares(2) == [1, 4]
assert squares(10) == [1, 4, 9, 16, 25, 36, 49, 64, 81, 100]
assert squares(11) == [1, 4, 9, 16, 25, 36, 49, 64, 81, 100, 121]
from pedal import *
from pedal.environments.nbgrader import setup_environment

if setup_environment('squares'):
    prevent_ast("If")
    prevent_ast("If")
    prevent_embedded_answer(1)
    prevent_import('numpy')


"""Check that squares raises an error for invalid inputs"""
try:
    squares(0)
except ValueError:
    pass
else:
    raise AssertionError("did not raise")

try:
    squares(-4)
except ValueError:
    pass
else:
    raise AssertionError("did not raise")
''', f.getvalue())

    def test_check_file(self):
        with io.StringIO() as f, redirect_stdout(f):
            nbgrader = setup_environment("*", main_file=here + 'datafiles/jupyter_notebook_example.ipynb')
            with self.assertRaises(AssertionError) as ae:
                prevent_ast("If")
            self.assertEqual("You used an IF statement on line 7. You may not use that.", str(ae.exception))



"""
Some kind of function to break up the sections
"""
import re
import sys
from html.parser import HTMLParser

from pedal.resolvers.core import make_resolver
from pedal.source import verify, next_section as original_next_section
from pedal.core.feedback import Feedback
from pedal.core.environment import Environment
from pedal.core.report import MAIN_REPORT
from pedal.sandbox import run, get_sandbox, set_input, start_trace
from pedal.source.sections import FeedbackSourceSection
from pedal.tifa import tifa_analysis
from pedal.resolvers.simple import resolve as simple_resolve, by_priority
from pedal.resolvers.sectional import resolve as original_sectional_resolve
from pedal.core.formatting import Formatter
from pedal.core.location import Location
from pedal.utilities.text import inject_line

try:
    import tabulate
except ImportError:
    tabulate = None


class VPLEnvironment(Environment):
    """
    Configures the BlockPy programming environment.
    """
    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='on_run.py', skip_tifa=True, skip_run=False,
                 inputs=None, set_correct=True, set_success=None,
                 report=MAIN_REPORT, trace=True):
        super().__init__(files=files, main_file=main_file, main_code=main_code,
                         user=user, assignment=assignment, course=course,
                         execution=execution, instructor_file=instructor_file,
                         report=report)
        self.skip_run = skip_run
        self.skip_tifa = skip_tifa
        self.trace = trace
        report.set_formatter(VPLFormatter(report))
        verify(report=self.report)
        if not skip_tifa:
            tifa_analysis(report=self.report)
        if inputs:
            set_input(inputs)
        if skip_run:
            student = get_sandbox(report=report)
        else:
            if trace:
                start_trace()
            student = run(report=report)
        self.fields = {
            'student': student,
            'resolve': resolve,
            'next_section': self.next_section,
            'sectional_resolve': sectional_resolve
        }

    def next_section(self, name=""):
        original_next_section(name=name, report=self.report)
        verify(report=self.report)
        if not self.skip_tifa:
            tifa_analysis(report=self.report)
        student = get_sandbox(report=self.report)
        if self.skip_run:
            student.clear()
        else:
            if self.trace:
                start_trace()
            student.clear()
            student = student.run()
        return student

    def load_main(self, path):
        """
        Allowed to return either a string value (the contents of the file)
        or the exception that was raised.
        """
        try:
            return super().load_main(path)
        except OSError as file_error:
            return file_error


setup_environment = VPLEnvironment

## TODO: Check if file exists

class VPLFormatter(Formatter):

    def pre(self, text):
        return "\n".join([">" + line for line in text.split("\n")])

    def python_code(self, code, focus=None):
        if isinstance(focus, Location):
            code = inject_line(code, focus.line + 1, " " * (focus.col - 1) + "^" * (focus.end_col - focus.col))
        return self.pre(code)

    def python_expression(self, code):
        return code

    def filename(self, filename):
        return filename

    def python_value(self, code):
        return self.pre(code)

    def inputs(self, inputs):
        return self.pre(inputs)

    def output(self, output):
        return self.pre(output)

    def traceback(self, traceback):
        return traceback

    def name(self, name):
        return name

    def line(self, line_number):
        return line_number

    def frame(self, name):
        return name

    def exception(self, exception):
        return self.pre(exception)

    def table(self, rows, columns):
        if tabulate:
            return self.pre(tabulate.tabulate(rows, headers=columns))
        else:
            return self.pre(f"Table:\n{columns}\n{rows}")


score_maximum = 1


def set_maximum_score(number):
    """
    TODO: Attach data to the Report instead of a global variable.

    Args:
        number: The maximum score to give.
    """
    global score_maximum
    score_maximum = number


@make_resolver
def resolve(report=MAIN_REPORT, priority_key=by_priority):
    """

    Args:
        report:
        custom_success_message:
    """
    print("<|--")
    final = simple_resolve(report, priority_key=priority_key)
    if final.positives:
        print("-Positive Notes")
        for positive in final.positives:
            print(positive)
    if final.systems:
        print("-System Notes")
        for system in final.systems:
            if system.label and system.label != "log":
                print(system.title)
            if system.label and system.label != "log":
                print(system.label)
            if system.message:
                print(system.message)
    print("-"+final.title)
    print(final.message)
    print("--|>")
    print("Grade :=>>", round(final.score*score_maximum))


import sys
stdout = sys.stdout

@make_resolver
def sectional_resolve(report=MAIN_REPORT, priority_key=by_priority):
    """

    Args:
        report:
        custom_success_message:
    """
    print("<|--")
    print(report.get_current_group(), file=stdout)
    finals = original_sectional_resolve(report, priority_key=priority_key)
    #if final.positives:
    #    print("-Positive Notes")
    #    for positive in final.positives:
    #        print(positive)
    global_feedbacks = sorted([f for f in report.feedback if f.parent is None],
                              key=lambda f: f.section_number if isinstance(f, FeedbackSourceSection) else -1)
    for global_feedback in global_feedbacks:
        if isinstance(global_feedback, FeedbackSourceSection):
            print(f"-Part {global_feedback.section_number}")
            final = finals.get(global_feedback)
            if final:
                print(final.title)
                print(final.message)
            else:
                print("No feedback for this section")
        # TODO: Should we be skipping all global feedback? Need more examples
        #else:
        #    print(global_feedback.title)
        #    print(global_feedback.message)
    print("--|>")
    total_score = sum(f.score for f in finals.values())
    print("Grade :=>>", round(total_score*score_maximum))



"""
Feedback functions of the Source module
"""
import traceback as tb
from pedal.core.commands import feedback
from pedal.core.feedback import FeedbackResponse
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.utilities.exceptions import ExpandedTraceback, get_exception_name
from pedal.source.constants import TOOL_NAME
from pedal.utilities.text import add_indefinite_article
from pedal.core.formatting import wrap_fields


class SourceFeedback(FeedbackResponse):
    """ Base class of all Feedback functions for Source Tool """
    category = feedback.CATEGORIES.SYNTAX
    kind = feedback.KINDS.MISTAKE
    tool = TOOL_NAME


class blank_source(SourceFeedback):
    """ Source code file was blank. """
    title = "No Source Code"
    message_template = "Source code file is blank."
    justification = "After stripping the code, there were no characters."


class not_enough_sections(SourceFeedback):
    """ Didn't have all the needed sections. """
    title = "Not Enough Sections"
    message_template = ("Tried to advance to next section but the "
                        "section was not found. Tried to load section "
                        "{count}, but there were only {found} sections.")
    justification = ("Section index exceeded the length of the separated "
                     "sections list.")

    def __init__(self, section_number, found, **kwargs):
        fields = {'count': section_number, 'found': found}
        super().__init__(fields=fields, **kwargs)


class source_file_not_found(SourceFeedback):
    """ No source file was given. """
    title = 'Source File Not Found'
    message_template = ("The given filename {name:filename} was either not "
                        "found or could not be opened. Please make sure the "
                        "file is available.")
    version = '0.0.1'
    justification = "IOError while opening file to set_source"

    def __init__(self, name, sections, **kwargs):
        report = kwargs.get("report", MAIN_REPORT)
        fields = {'name': name, 'sections': sections}
        group = 0 if sections else kwargs.get('group')
        super().__init__(fields=fields, group=group, **kwargs)


class syntax_error(SourceFeedback):
    """ Generic feedback for any kind of syntax error. """
    muted = False
    title = "Syntax Error"
    constant_fields = {"suggestion": "Suggestion: Check line {lineno:line}, the line before it, and the line "
                                     "after it. Ignore blank lines."}
    message_template = ("Bad syntax on line {lineno:line}.\n\n"
                        "{traceback_preamble}{traceback_message}\n"
                        "{exception_message}"
                        "{suggestion_message}")
    version = '1.0.0'
    justification = "Syntax error was triggered while calling ast.parse"

    def __init__(self, line, filename, code, col_offset,
                 exception, exc_info, enhance=True, **kwargs):
        report = kwargs.get('report', MAIN_REPORT)
        files = report.submission.get_files_lines()
        if filename not in files:
            files[filename] = code.split("\n")
        if report.submission is not None:
            lines = report.submission.get_lines()
            line_offsets = report.submission.line_offsets
        else:
            lines = code.split("\n")
            line_offsets = {}
        exception_name = get_exception_name(exception)
        exception_name_proper = add_indefinite_article(exception_name)
        traceback = ExpandedTraceback(exception, exc_info, False,
                                      [report.submission.instructor_file],
                                      line_offsets, [filename], lines, files)
        traceback_stack = traceback.build_traceback()
        traceback_message = traceback.format_traceback(traceback_stack, report.format)
        traceback_preamble = f"The traceback was:\n" if traceback_message else ""
        #if not enhance:
        #    self.message_template = "{traceback_message}\n{exception_message}"
        line_offset = line_offsets.get(filename, 0)
        exception_message = self.make_exception_message(exception)
        fields = {'lineno': line + line_offset,
                  'filename': filename,
                  'offset': col_offset,
                  'exception': exception,
                  'exception_name': exception_name,
                  'exception_name_proper': exception_name_proper,
                  'exception_message': exception_message,
                  'traceback': traceback,
                  'traceback_stack': traceback_stack,
                  'traceback_preamble': traceback_preamble,
                  'traceback_message': traceback_message}
        location = Location(line=line + line_offset, col=col_offset, filename=filename)
        fields['suggestion_message'] = self.constant_fields['suggestion'].format(**wrap_fields(report.format, fields))
        super().__init__(fields=fields, location=location, **kwargs)

    def make_exception_message(self, exception):
        """ Generate a suggestion message based on the exception. """
        base_message = exception.msg
        base_message = base_message.capitalize() + "."
        #if base_message == "Invalid syntax.":
        #    return ""
        return base_message + "\n"


class incorrect_number_of_sections(SourceFeedback):
    """ Incorrect number of sections """
    title = "Incorrect Number of Sections"
    message_template = ("Incorrect number of sections in your file. "
                        "Expected {count}, but only found {found}")
    justification = ""

    def __init__(self, count, found, **kwargs):
        fields = {'count': count, 'found': found}
        super().__init__(fields=fields, **kwargs)


# TODO: TabError


class indentation_error(syntax_error):
    """ Generic feedback for indentation errors. """
    title = "Indentation Error"
    constant_fields = {'suggestion': "Suggestion: Check line {lineno:line}, the line before it, and the "
                                     "line after it. Ignore blank lines."}
    message_template = ("Bad indentation on line {lineno:line} or adjacent line.\n\n"
                        "{traceback_preamble}{traceback_message}\n"
                        "{exception_message}"
                        "{suggestion_message}"
                        )
    version = '1.0.0'
    justification = "Indentation error was triggered while calling ast.parse"



from pedal.types.new_types import ModuleType, FunctionType, void_function, ClassType, \
    InstanceType, bool_function, int_function, str_function, register_builtin_module, AnyType, ListType

# TODO: route, Button, ...

_VOID_FUNCTIONS = ['start_server', 'show_debug_information', 'hide_debug_information']
_COMPONENT_FUNCTIONS = ["Image", "Link", "TextBox", "TextArea", "SelectBox",
                        "CheckBox", "LineBreak", "HorizontalRule", "Button",
                        "NumberedList", "BulletedList", "Header", "Table"]
_IDENTITY_FUNCTIONS = ["route"]


def build_drafter_module():
    _PageObject = ClassType('Page', {
        'state': AnyType(),
        'content': ListType(AnyType())
    }, [])

    def _PageObject_Constructor():
        return InstanceType(_PageObject)

    _ComponentObject = ClassType('Component', {}, [])

    def _ComponentObject_Constructor():
        return InstanceType(_ComponentObject)

    _DRAFTER_FIELDS = {
        'Page': FunctionType('Page', returns=_PageObject_Constructor),
    }

    _DRAFTER_FIELDS.update({
        name: void_function(name) for name in _VOID_FUNCTIONS
    })
    _DRAFTER_FIELDS.update({
        name: FunctionType(name, returns=_ComponentObject_Constructor)
        for name in _COMPONENT_FUNCTIONS
    })
    _DRAFTER_FIELDS.update({
        name: FunctionType(name, returns='identity')
        for name in _IDENTITY_FUNCTIONS
    })
    return ModuleType('drafter', _DRAFTER_FIELDS)


register_builtin_module('drafter', build_drafter_module)



"""
Tests related to using alternative testing libraries
"""

import unittest

from pedal import unit_test, block_function, set_source, clear_report
from pedal.assertions.static import *
from pedal.assertions.commands import *
from pedal.assertions.runtime import *
from pedal.resolvers import simple
from pedal.types.new_types import DictType, LiteralStr, StrType
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE


STUDENT_TEST_CODE_DIVIDE = """
import unittest

def divide(a, b):
    return a / b

class TestDivide(unittest.TestCase):
    def test_divide_success(self):
        self.assertEqual(divide(1, 1), 1)
    def test_divide_error(self):
        self.assertEqual(divide(1, 0), 0)
    def test_divide_failure(self):
        self.assertEqual(divide(1, 2), 2)
"""

class TestUnittest(ExecutionTestCase):
    def test_unittest_library(self):
        clear_report()
        set_source(STUDENT_TEST_CODE_DIVIDE)
        run()
        ensure_unittest_tests()
        final = simple.resolve()
        self.assertEqual("""2/3 of your unit tests are not passing.""", final.message.rstrip())

    def test_unittest_library_explicit(self):
        clear_report()
        set_source(STUDENT_TEST_CODE_DIVIDE+'\nunittest.main(exit=False)')
        run()
        ensure_unittest_tests()
        final = simple.resolve()
        self.assertEqual("""2/3 of your unit tests are not passing.""", final.message.rstrip())

class TestPyTest(ExecutionTestCase):
    def test_pytest_library(self):
        clear_report()
        set_source("""import pytest
def divide(a, b):
    return a / b

def test_divide_success():
    assert divide(1, 1) == 1
def test_divide_error():
    assert divide(1, 0) == 0
def test_divide_failure():
    assert divide(1, 2) == 2""")
        run()
        ensure_pytest_tests()
        final = simple.resolve()
        self.assertEqual("""2/3 of your unit tests are not passing.""", final.message.rstrip())

if __name__ == '__main__':
    unittest.main(buffer=False)





"""
File containing errors throwable by Pedal.
"""

__all__ = ['PedalToolNotRegistered', 'PedalToolAlreadyRegistered']


class PedalToolNotRegistered(Exception):
    """ Raised when you reference a Tool that is not registered. """
    def __init__(self, missing_name, available_names):
        message = (f"The tool {missing_name!r} is not registered. "
                   f"Available tools are:\n{available_names!r}")
        super().__init__(message)


class PedalToolAlreadyRegistered(Exception):
    """ Raised when you register an already-existing tool. """
    def __init__(self, existing_name):
        message = f"The tool {existing_name!r} is already registered."
        super().__init__(message)



"""
Module for the CaitNode class, which wraps AST nodes.
"""
import ast
from pedal.cait.ast_helpers import dump
from types import MethodType

from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.utilities.system import IS_AT_LEAST_PYTHON_38, IS_SKULPT


class CaitNode:
    """
    A wrapper class for AST nodes. Linearizes access to the children of the ast
    node and saves the field this AST node originated from.

    Attributes:
        ast_node (ast.AST): The original AstNode.

    TODO: May want to just add fields and methods to the existing AST nodes and
    use a production pattern instead.
    """

    def __init__(self, ast_node, my_field='', tid=0, lin_tree=None,
                 ancestor=None, report=None):
        """

        Args:
            ast_node (ast_node): The AST node to be wrapped
            my_field (str): the field of the parent node that produced this child.
            tid (int): the tree id
            lin_tree list of cait_node: A linear version of the tree
            ancestor (cait_node): The parent of this node
            report: The report associated with this particular match.
        """
        if report is None:
            report = MAIN_REPORT
        self.report = report
        self.children = []
        self.astNode = ast_node
        self.field = my_field
        self.tree_id = tid
        self.parent = ancestor
        if lin_tree is None:
            self.linear_tree = [self]
        else:
            lin_tree.append(self)
            self.linear_tree = lin_tree

        # reference to the easy node wrapping the ast_node
        setattr(ast_node, 'cait_node', self)

        tid_count = tid

        my_field_generator = ast.iter_fields(self.astNode)
        for item in my_field_generator:
            field, value = item
            # if the field doesn't have a value, no child exists
            if value is None:
                continue

            # If the children are not in an array, wrap it in an array for
            # consistency in the code the follows
            if not isinstance(value, list):
                value = [value]

            # Reference ast_node_visitor.js for the original behavior and keep note of it for the purposes of handling
            # the children noting the special case when the nodes of the array are actually parameters of the node
            # (e.g. a load function) instead of a child node
            for sub_value in value:
                if isinstance(sub_value, ast.AST):
                    new_child = CaitNode(sub_value, my_field=field,
                                         tid=tid_count + 1,
                                         lin_tree=self.linear_tree,
                                         ancestor=self,
                                         report=self.report)
                    self.children.append(new_child)
                    tid_count = len(self.linear_tree) - 1

    def __str__(self):
        return ''.join([self.field, "\n", dump(self.astNode)])

    def numeric_logic_check(self, mag, expr):
        """
        If this node is a Compare or BoolOp node, sees if the logic in expr (a javascript string being a logical
        statement) matches the logic of self. This assumes that we are only comparing numerical values to a single
        variable
        TODO: modify this to take multiple variables
        TODO: modify to support more than +, -, *, and / BinOps
        TODO: modify to support unary operators other than USub and Not
        TODO: This is very finicky and buggy, try not to use it
        Args:
            mag (float): the order of magnitude that should be added to numbers to check logic, 1 is usually a good value,
                    especially when working with the set of integers.
            expr (Compare or BoolOp): the "Compare" or "BoolOp" tree to check self against

        Returns:
            bool: True if self (typically student node) and expr are equivalent boolean expressions
        """

        def eval_unop(unop_num, unop_node):
            """

            Args:
                unop_num:
                unop_node:

            Returns:

            """
            operand = eval_selector(unop_num, unop_node.operand)
            op = unop_node.op_name

            return {"USub": -operand,
                    "Not": not operand}[op]

        def eval_binop(binop_num, binop_node):
            """

            Args:
                binop_num:
                binop_node:

            Returns:

            """
            left = eval_selector(binop_num, binop_node.left)
            right = eval_selector(binop_num, binop_node.right)
            op = binop_node.op_name

            return {
                "Add": left + right,
                "Sub": left - right,
                "Mult": left * right,
                "Div": left / right}[op]

        def eval_selector(op_num, op_expr):
            """

            Args:
                op_num:
                op_expr:

            Returns:

            """
            op_expr = op_num if op_expr.ast_name == "Name" else op_expr
            if isinstance(op_expr, (int, float)):
                return op_expr
            if op_expr.ast_name == "BinOp":
                return eval_binop(op_num, op_expr)
            if op_expr.ast_name == "UnaryOp":
                return eval_unop(op_num, op_expr)
            if op_expr.ast_name == "Num":
                return op_expr.n
            if op_expr.ast_name == "Constant":
                return op_expr.value
            raise NotImplementedError

        def eval_bool_comp(num_list, comp_ast):
            """

            Args:
                num_list:
                comp_ast:

            Returns:

            """
            ops = comp_ast.ops_names
            comps = comp_ast.comparators
            results = []
            current = comp_ast.left
            left = current

            for num_i in num_list:
                result = True
                for op, comp in zip(ops, comps):
                    current = eval_selector(num_i, current)
                    comp_p = eval_selector(num_i, comp)

                    res = {
                        "Eq": current == comp_p,
                        "NotEq": current != comp_p,
                        "Lt": current < comp_p,
                        "LtE": current <= comp_p,
                        "Gt": current > comp_p,
                        "GtE": current >= comp_p,
                    }[op]
                    current = comp
                    result = result and res
                    if not result:
                        break
                results.append(result)
                current = left
            return results

        def eval_boolop(num_list, boolop_ast):
            """

            Args:
                num_list:
                boolop_ast:

            Returns:

            """
            boolop = boolop_ast.op_name
            values = boolop_ast.values
            results_c = None
            is_and = boolop == "And"
            for value in values:
                if value.ast_name == "Compare":
                    results = eval_bool_comp(num_list, value)
                else:  # should be boolop
                    results = eval_boolop(num_list, value)
                if results_c is None:
                    results_c = results
                else:  # compile results
                    new_result = []
                    for result1, result2 in zip(results_c, results):
                        if is_and:
                            new_result.append(result1 and result2)
                        else:
                            new_result.append(result1 or result2)
                    results_c = new_result
            return results_c

        try:
            ins_expr = CaitNode(ast.parse(expr), report=self.report).body[0].value
            ins_nums = ins_expr.find_all(["Num", "Constant"])
            std_nums = self.find_all(["Num", "Constant"])
            test_nums = []
            for num in ins_nums:
                if hasattr(num, "n"):
                    raw_num = num.n
                else:
                    raw_num = num.value
                test_nums.append(raw_num)
                test_nums.append(raw_num + mag)
                test_nums.append(raw_num - mag)
            for num in std_nums:
                if hasattr(num, "n"):
                    raw_num = num.n
                else:
                    raw_num = num.value
                test_nums.append(raw_num)
                test_nums.append(raw_num + mag)
                test_nums.append(raw_num - mag)

            if self.ast_name == "Compare":
                std_res = eval_bool_comp(test_nums, self)
            elif self.ast_name == "BoolOp":
                std_res = eval_boolop(test_nums, self)
            else:
                return False

            if ins_expr.ast_name == "Compare":
                ins_res = eval_bool_comp(test_nums, ins_expr)
            elif ins_expr.ast_name == "BoolOp":
                ins_res = eval_boolop(test_nums, ins_expr)
            else:
                raise TypeError
            return ins_res == std_res
        except Exception:
            return False

    def get_next_tree(self):
        """Gets the next tree in the AST
        This method gets the next AST node that is of equal or higher level than self. Returns None if the end of the
        tree is reached
        TODO: Create a get sibling method.

        Returns:
            cait_node: The next tree in the AST

        """

        # adding function to track tree ids
        def visit_counter(self, node):
            """

            Args:
                self:
                node:
            """
            self.counter += 1
            self.generic_visit(node)

        node_counter = ast.NodeVisitor()
        setattr(node_counter, 'counter', self.tree_id)
        node_counter.visit = MethodType(visit_counter, node_counter)

        # getting ids
        node_counter.visit(self.astNode)
        out_of_tree = node_counter.counter >= len(self.linear_tree)  # check if out of bounds
        # len(self.children) > 0 and self.children[-1] == node_counter
        if out_of_tree:
            return None
        return self.linear_tree[node_counter.counter]

    def get_child(self, node):
        """

        Args:
            node: a non-CaitNode ast node

        Returns:
            cait_node: the corresponding cait_node to the child
        """
        if isinstance(node, ast.AST):
            for child in self.children:
                if child.astNode == node:
                    return child
        elif isinstance(node, int):
            return self.children[node]
        return None

    @staticmethod
    def get_ast_name(node):
        """

        Args:
            node:

        Returns:

        """
        return type(node).__name__

    def get_clashing_attr(self, key):
        """

        Args:
            key:

        Returns:

        """
        if key == "value":
            return self.get_value()

    def __getattr__(self, item):
        key = item
        """
        Non-ast node attributes based on ast_node attributes
        """
        node_name = CaitNode.get_ast_name(self.astNode)
        if node_name == "Assign" and key == "target":
            key = "targets"
        if item in AST_SINGLE_FUNCTIONS:
            key = item[:-5]  # strip suffix '_name'
        if item in AST_ARRAYS_OF_FUNCTIONS:
            key = item[:-6]  # strip suffix '_names'

        """
        Non-ast node attributes
        """
        if key == 'next_tree':
            return self.get_next_tree()
        if key == 'ast_name':
            return node_name
        elif key == '_name':
            return self.astNode.name
        elif key == 'ast_node':
            return self.astNode
        else:  # ast node attributes or derivative attributes
            if hasattr(self.astNode, key):
                # noinspection PyBroadException
                try:
                    field = self.astNode.__getattribute__(key)
                except Exception:
                    field = None
                if node_name == "Assign" and item != key:
                    if item == "target":
                        return field[0].cait_node  # Get's the relevant ast node
                    elif item == "targets" and isinstance(field, list):
                        easy_array = []
                        for node in field:
                            easy_array.append(node.cait_node)
                        return easy_array
                    else:
                        return field
                elif item in AST_SINGLE_FUNCTIONS:
                    return type(field).__name__
                elif item in AST_ARRAYS_OF_FUNCTIONS:
                    str_ops_list = []
                    for op in field:
                        str_ops_list.append(type(op).__name__)
                        return str_ops_list
                elif isinstance(field, ast.AST):
                    return field.cait_node
                elif isinstance(field, list):
                    try:
                        return [f.cait_node for f in field]
                    except AttributeError:
                        # This can only happen in NonLocals, which has a list
                        # of raw strings in the `names` property
                        return field
                else:
                    return field
            else:  # get added field that may have existed for different node types
                return self.get_clashing_attr(key)

    def find_matches(self, pattern, is_mod=False, check_meta=True, use_previous=True):
        """
        Retrieves any patterns that match against this CaitNode. Expected to be
        used for subpattern matching.
        Args:
            pattern (CaitNode/str): The pattern searched for in this cait_node
            is_mod (bool): Assumes that this node is a module node
            check_meta (bool): Checks if meta information of the nodes also matches (e.g. instead of just checking children,
                            also checking whether the children have the same field name in both the pattern and the
                            source
            use_previous (bool): If True, the match will be searched while inheriting the symbol table of the parent. This
                            means that variable consistency must be maintained between parent and child matches.

        Returns:
            :obj: 'list' of :'obj': AstMap: a list of matches

        """
        # Avoid circular import
        import pedal.cait.stretchy_tree_matching as stm
        is_node = isinstance(pattern, CaitNode)
        if not isinstance(pattern, str) and not is_node:
            raise TypeError("pattern expected str or CaitNode, found {0}".format(type(pattern)))
        matcher = stm.StretchyTreeMatcher(pattern, report=self.report)
        if (not is_node and not is_mod) and len(matcher.root_node.children) != 1:
            raise ValueError("pattern does not evaluate to a singular statement")
        use_previous = self.map if use_previous else None
        return matcher.find_matches(self, check_meta=check_meta, use_previous=use_previous)

    def find_match(self, pattern, is_mod=False, check_meta=True, use_previous=True):
        """

        Args:
            pattern:
            is_mod:
            check_meta:
            use_previous:

        Returns:

        """
        matches = self.find_matches(pattern, is_mod, check_meta=check_meta, use_previous=use_previous)
        if len(matches) != 0:
            return matches[0]
        return None

    def find_all(self, node_type):
        """Finds all nodes defined by string node_type

        Args:
            node_type: the string representing the "type" of node to look for

        Returns:
            a list of Ast Nodes (cait_nodes) of self that are of the specified type (including self if self
                    meets that criteria)
        """
        items = []
        visitor = ast.NodeVisitor()
        setattr(visitor, "items", items)

        def main_visit(self, node):
            """

            Args:
                self:
                node:

            Returns:

            """
            self.items.append(node.cait_node)
            return self.generic_visit(node)

        if type(node_type) is not list:
            node_type_list = [node_type]
        else:
            node_type_list = node_type

        for node_t in node_type_list:
            if (IS_AT_LEAST_PYTHON_38 and not IS_SKULPT) and node_t in ('Num', 'Str', 'Bool'):
                func_name = 'visit_Constant'
                func_ref = self._handle_visit_constant(node_t, main_visit)
            else:
                func_name = 'visit_' + node_t
                func_ref = main_visit
            setattr(visitor, func_name, MethodType(func_ref, visitor))

        visitor.visit(self.astNode)
        return visitor.items

    def _handle_visit_constant(self, actual_node, visitor_function):
        def visit_Constant(self, node):
            # TODO: Does this need to be extended for None, tuple, frozenset, anything else?
            if actual_node == 'Bool' and isinstance(node.value, bool):
                return visitor_function(self, node)
            if actual_node == 'Num' and isinstance(node.value, (int, float)) and not isinstance(node.value, bool):
                return visitor_function(self, node)
            elif actual_node == 'Str' and isinstance(node.value, str):
                return visitor_function(self, node)
            return self.generic_visit(node)
        return visit_Constant

    def has(self, node):
        """
        Determine if this node has the given `node`. Specifically, it checks if the ast has a given Constant or variable
        This method does NOT check for equality of two asts. Use find_matches for that functionality
        Args:
            node (): A constant (int, float, str), or Name astnode

        Returns: True if found, False otherwise

        """
        if isinstance(node, (int, float, str)):
            visitor = ast.NodeVisitor()
            has_constant = []

            def visit_Constant(self, potential):
                has_constant.append(node == potential.value)
                return self.generic_visit(potential)

            def visit_Num(self, potential):
                has_constant.append(node == potential.n)
                return self.generic_visit(potential)

            def visit_Str(self, potential):
                has_constant.append(node == potential.s)
                return self.generic_visit(potential)

            visitor.visit_Num = MethodType(visit_Num, visitor)
            visitor.visit_Constant = MethodType(visit_Constant, visitor)
            visitor.visit_Str = MethodType(visit_Str, visitor)
            visitor.visit(self.astNode)
            return any(has_constant)
        elif node.ast_name != "Name":
            return False
        visitor = ast.NodeVisitor()
        has_name = []

        def visit_Name(self, potential):
            """

            Args:
                self:
                potential:

            Returns:

            """
            has_name.append(node.id == potential.id)
            return self.generic_visit(potential)

        visitor.visit_Name = MethodType(visit_Name, visitor)
        visitor.visit(self.astNode)
        return any(has_name)

    def is_before(self, other):
        """
        Uses tree id to check if self node came before other.
        Args:
            other (cait_node): the other node to compare to

        Returns:
            bool: True if self is before other
        """
        try:
            return self.tree_id < other.tree_id and self.linear_tree == other.linear_tree
        except Exception:
            raise TypeError

    def is_ast(self, ast_name):
        """
        Checks self is the type of the specified ast node
        Args:
            ast_name (str): The name of the ast node type

        Returns:
            bool: True if this node's ast name matches the specified one
        """
        # TODO: Deal with this compatibility issue in a more elegant way
        if not isinstance(ast_name, str):
            ast_name = CaitNode.get_ast_name(ast_name.astNode)
        self_name = CaitNode.get_ast_name(self.astNode).lower()
        other_name = ast_name.lower()
        evaluation = self_name == other_name
        if self_name == "constant" and other_name in ["num", "str", "bytes"]:
            if other_name == "num" and type(self.n) in [float, int]:
                evaluation = True
            elif other_name == "str" and type(self.s) is str:
                evaluation = True
            elif other_name == "bytes" and type(self.s) is bytes:
                evaluation = True
        return evaluation

    def is_method(self):
        """
        Checks if self is a method

        Returns:
            bool: True if I'm a FunctionDef, and if any of my parents are ClassDef.
        """
        # Check if I'm a FunctionDef, and if any of my parents are ClassDef.
        if self.ast_name != "FunctionDef":
            return False
        current = self.parent
        while current is not None:
            if current.ast_name == "ClassDef":
                return True
            # Don't treat closures as methods
            elif current.ast_name == "FunctionDef":
                return False
            current = current.parent
        return False

    def get_data_state(self):
        """
        Gets the data_state object of self

        Returns:
            data_state or None: returns data_state if self is a name and exists, otherwise None
        """
        if self.ast_name != "Name":
            return None
        try:
            return self.report['tifa']['latest'].top_level_variables[self.id]
        except KeyError:
            return None

    def get_data_type(self):
        """

        Returns:
            type of the variable associated with this node if it's a name node, otherwise None.
        """
        state = self.get_data_state()
        if state is None:
            return None
        else:
            return state.type

    def was_type(self, tp):
        """

         Returns:
             type of the variable associated with this node if it's a name node, otherwise None.
         """
        state = self.get_data_state()
        if state is None:
            return None
        else:
            return state.was_type(tp)

    def get_value(self):
        """"
        Returns:
            Value of node if Num or Str, and get_data_state if Name
        """
        value = None
        if self.is_ast("Num"):
            value = self.n
        elif self.is_ast("Str"):
            value = self.s
        elif self.is_ast("Constant"):
            value = self.value
        elif self.is_ast("Name"):
            # TODO: Decide on what this should return...
            value = self.id
        return value

    def locate(self):
        """
        Returns:
            pedal.core.Location: The location of this node.
        """
        return Location(self.astNode.lineno, col=self.astNode.col_offset)

    def to_json(self):
        return str(self)


AST_SINGLE_FUNCTIONS = ["ctx_name", "op_name"]
AST_ARRAYS_OF_FUNCTIONS = ["ops_names"]



"""
Utilities for handling exceptions across all of Pedal.
"""
from textwrap import indent
import traceback
import os
import sys

from pedal.utilities.system import IS_AT_LEAST_PYTHON_310, IS_AT_LEAST_PYTHON_311, IS_SKULPT
from pedal.core.location import Location

BuiltinKeyError = KeyError

_backup_stdout = sys.stdout


class KeyError(BuiltinKeyError):
    """
    A version of KeyError that replaces the built-in with one small
    modification: when printing an explanatory message, the message is not
    rendered as a tuple. Because that's stupid and the fact that it made it
    into CPython is just rude.

    See Also:
        https://github.com/python/cpython/blob/master/Objects/exceptions.c#L1556
    """
    __module__ = "builtins"

    def __init__(self, original, message):
        for field in ['__cause__', '__traceback__', '__context__']:
            if hasattr(original, field):
                setattr(self, field, getattr(original, field))
            else:
                setattr(self, field, None)
        self.message = message

    def __str__(self):
        return self.message


def improve_builtin_exceptions(exception):
    """
    Modify the given builtin exception to have a better interface.
    Currently used to make KeyError look like other exceptions.

    Args:
        exception (Exception): The existing exception to modify.

    Returns:
        Exception: A new exception, or the original one unchanged.
    """
    if isinstance(exception, BuiltinKeyError):
        return KeyError(exception, "key not found")
    return exception


def get_exception_name(exception: Exception) -> str:
    """
    Gets the name of the exception (e.g., IndexError gives ``"IndexError"``).
    """
    return exception.__class__.__name__


def add_context_to_error(e, message):
    """

    Args:
        e:
        message:

    Returns:

    """
    if isinstance(e, BuiltinKeyError):
        new_args = repr(e.args[0]) + message
        e = KeyError(e, new_args)
        e.args = tuple([new_args])
    elif isinstance(e, OSError):
        # TODO: Investigate OSError, since they have so many args.
        #       Might be weird.
        e.args = tuple([e.args[0] + message])
        return e
    elif hasattr(e, 'args') and e.args:
        e.args = tuple([e.args[0] + message])
    return e


class FakeFrame:
    """ Simplistic hack to make fake StackFrames for SyntaxErrors """
    def __init__(self, name, filename, lineno, line, colno, end_lineno, end_colno):
        self.name = name
        self.filename = filename
        self.lineno = lineno
        self._line = line
        self.colno = colno
        self.end_lineno = end_lineno
        self.end_colno = end_colno

    @property
    def line(self):
        return self._line

    def __eq__(self, other):
        try:
            return self.filename == other.filename and self.lineno == other.lineno
        except:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class ExpandedTraceback:
    """
    Class for reformatting tracebacks to have more pertinent information.
    """

    def __init__(self, exception, exc_info, full_traceback,
                 hide_filenames, line_offsets, show_filenames,
                 original_code_lines, student_files):
        """
        Args:
            exception (Exception): The exception that was raised.
            exc_info (ExcInfo): The result of sys.exc_info() when the exception
                was raised.
            full_traceback (bool): Whether or not to provide the full traceback
                or just the parts relevant to students.
        """
        self.line_offsets = line_offsets
        self.exception = exception
        self.exc_info = exc_info
        self.full_traceback = full_traceback
        self.hide_filenames = hide_filenames
        self.show_filenames = show_filenames
        self.line_number = traceback.extract_tb(exc_info[2])[-1][1]
        self.original_code_lines = original_code_lines
        self.student_files = student_files

    def __repr__(self):
        return f"ExpandedTraceback({self.exception})"

    def build_traceback(self):
        """
        Filter out unnecessary frames
        """
        if not self.exception:
            return []
        cl, exc, tb = self.exc_info
        while tb and self._is_relevant_tb_level(tb):
            tb = tb.tb_next
        length = self._count_relevant_tb_levels(tb)
        tb_e = traceback.TracebackException(cl, self.exception, tb, limit=length,
                                            capture_locals=False)
        for frame in tb_e.stack:
            self._fix_frame_line(frame)
        frames = list(tb_e.stack)
        # A SyntaxError has to be handled differently to actually get its output:
        # https://docs.python.org/3/library/traceback.html#traceback.print_exception
        if isinstance(self.exception, SyntaxError):
            offset = self.exception.offset
            if IS_AT_LEAST_PYTHON_310 and not IS_SKULPT:
                end_lineno = self.exception.end_lineno
                end_offset = offset if self.exception.end_offset not in {None, 0} else offset
                end_offset = offset + 1 if offset == end_offset or end_offset == -1 else end_offset
            else:
                end_lineno = self.exception.lineno
                end_offset = 1 + offset
            fake_frame = FakeFrame("<module>", self.exception.filename,
                                   self.exception.lineno, None, offset-1, end_lineno, end_offset-1)
            self._fix_frame_line(fake_frame)
            # Skulpt compatibility hack, to prevent duplicate tracebacks
            if not frames or fake_frame != frames[-1]:
                frames.append(fake_frame)
        return frames

    def _fix_frame_line(self, frame):
        if frame.filename in self.line_offsets:
            original_lineno = frame.lineno
            frame.lineno += self.line_offsets[frame.filename]
            if original_lineno - 1 < len(self.original_code_lines):
                frame._line = self.original_code_lines[original_lineno - 1]
            else:
                frame._line = "# *line missing*"
        else:
            # print(frame.filename, self.student_files, frame.lineno)
            if frame.filename in self.student_files:
                if frame.lineno - 1 < len(self.original_code_lines):
                    frame._line = self.student_files[frame.filename][frame.lineno - 1]
                else:
                    # Not actually possible in CPython, but Skulpt gives weird
                    #   SyntaxErrors that are technically the "next" line.
                    frame._line = ""

    def _count_relevant_tb_levels(self, tb):
        length = 0
        while tb and not self._is_relevant_tb_level(tb):
            length += 1
            tb = tb.tb_next
        return length

    def _is_relevant_tb_level(self, tb):
        """
        Determines if the give part of the traceback is relevant to the user.

        Returns:
            boolean: True means it is NOT relevant
        """
        # Are in verbose mode?
        if self.full_traceback:
            return False
        filename, a_, b_, _ = traceback.extract_tb(tb, limit=1)[0]
        # Is the error in the student file?
        if filename in self.show_filenames:
            return False
        # Is the error in the instructor file?
        if filename in self.hide_filenames:
            return True
        # Is the error in this test directory?
        current_directory = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        if filename.startswith(current_directory):
            return True
        # Is the error related to a file in the parent directory?
        # parent_directory = os.path.dirname(current_directory)
        # TODO: Currently we don't refer to this?
        # Is the error in a local file?
        if filename.startswith('.'):
            return False
        # Is the error in an absolute path?
        if not os.path.isabs(filename):
            return False
        # Okay, it's not a student related file
        return True

    def format_traceback(self, traceback_stack, formatter):
        """
        Turn a stack of tracebacks into a message.

        Args:
            traceback_stack (list[Frame]): The actual traceback to format into
                a message.
            formatter (:py:class:`pedal.core.formatters.Formatter`): The
                formatter to use to make the message.

        Returns:
            str: The text of the message.
        """

        # TODO: In 3.10 (or 11?), the colno is available. Indents are not automatically applied. So we need to have
        # different behavior for those versions to show off the entire line and also highlight "the good bit".
        traceback_message = "\n".join([
            (f"Line {formatter.line(frame.lineno)}"
             f" of file {formatter.filename(frame.filename)}" +
             (f" in {formatter.frame(frame.name)}\n"
              if frame.name != "<module>" and frame.name is not None
              else "\n") +
             f"{self.format_line(formatter, frame)}\n")
            for frame in traceback_stack
        ])
        if not traceback_message:
            return None
        return formatter.traceback(traceback_message)

    def format_line(self, formatter, frame):
        if IS_AT_LEAST_PYTHON_311:
            end_offset = frame.end_colno+1 if frame.lineno == frame.end_lineno else len(frame._line)
            # Note: Need to use _line because in 3.10 and above, the line gets stripped.
            # https://github.com/python/cpython/commit/5644c7b3ffd49bed58dc095be6e6148e0bb4431e
            line = frame._line if frame._line is not None else ''
            return formatter.python_code(line, focus=Location(0, frame.colno+1, 0, end_offset))
        elif IS_AT_LEAST_PYTHON_310:
            return formatter.python_code(frame._line if frame._line is not None else '')
        return formatter.python_code(frame.line if frame.line is not None else '')



"""
A completely barebones environment that loads nothing.
"""

from pedal.core.environment import Environment


class NoneEnvironment(Environment):
    """
    The NoneEnvironment is an environment that does pretty much nothing.
    It actually just subclasses the abstract Environment without extending
    any behavior.
    """


setup_environment = NoneEnvironment



def get_default_tifa_settings():
    return {
        # Type System
        'accept_generic_types': True,
        'numeric_type_equality': True,
        'struct_type': {
            'dataclasses': True,
            'typed_dict': True,
            'record_dict': True,
            'class': True
        },
        'enforce_nominal_types': True,
        'evaluate_string_literal_types': False,
        'allow_type_changes': True,
        'truthiness_returns_booleans': True,
        # Flow System
        'allow_global_writes': False,
        'allow_unused_variables': False,
        'allow_unused_return_value': False,
        'allow_unused_expression_value': False,
        # Bad Code
        'allow_redundant_equal_true': False,
        'allow_redundant_empty_else': False,
        'allow_unnecessary_if_return': False,
        # Versioning
        'type_system_version': 2
    }



import unittest
import ast
import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from pedal.cait.cait_api import *


class CaitNodeTest(unittest.TestCase):
    def test_init_(self):
        program = ast.parse("fun = 0 + 1\nfun01 =  2 + 3\nfun02 = 4 + 5")
        program = CaitNode(program)
        second_assign = program.children[1]
        third_assign = program.children[2]
        self.assertTrue(second_assign == program.linear_tree[second_assign.tree_id], "Nodes not being index properly")
        self.assertTrue(third_assign == program.linear_tree[third_assign.tree_id], "Nodes not being index properly-2")

    def test_get_next_tree(self):
        program = ast.parse("fun = 0 + 1\nfun01 =  2 + 3\nfun02 = 4 + 5")
        program = CaitNode(program)
        first_assign_c = program.children[0]
        second_assign_c = program.children[1]
        third_assign_c = program.children[2]
        second_assign_test = first_assign_c.get_next_tree()
        with self.subTest():
            self.assertTrue(second_assign_test == second_assign_c, "next_tree not found, when it should be")
        with self.subTest():
            no_next_end = third_assign_c.get_next_tree()
            self.assertTrue(no_next_end is None, "next_tree found when it shouldn't be at end")

    def test_find_all(self):
        program = ast.parse(
            "\nmy_list = [1, 2, 3]\nif True:\n\tmy_list.append(4)\nfor item in my_list:\n\tif item == 1:"
            "\n\t\titem += 1\n\tif item == 2:\n\t\titem += 4\n\tif item == 3:\n\t\titem += 6")
        program = CaitNode(program)
        if_set = program.find_all("If")
        self.assertTrue(len(if_set) == 4, "Found {} ifs, when 4 should be found".format(len(if_set)))

        num_set = program.find_all("Num")
        self.assertTrue(len(num_set) == 10, "Found {} nums, when 10 should be found".format(len(num_set)))

        num_if_set = program.find_all(["Num", "If"])
        self.assertTrue(len(num_if_set) == 14, "Found {} nums, when 10 should be found".format(len(num_if_set)))

        for_node = program.children[2]
        if_set_for = for_node.find_all("If")
        self.assertTrue(len(if_set_for) == 3, "Found {} ifs, when 3 should be found".format(len(if_set_for)))

        if0_node = program.children[1]
        if_set_if0 = if0_node.find_all("If")
        self.assertTrue(len(if_set_if0) == 1, "Found {} ifs, when 1 should be found".format(len(if_set_if0)))

    def test_has(self):
        program = ast.parse("x\n"
                            "y\n"
                            "x = 0\n"
                            "y = 'fun'")
        program = CaitNode(program)
        x = program.body[0].value
        y = program.body[1].value
        line3 = program.body[2]
        line4 = program.body[3]
        self.assertTrue(line3.has(x))
        self.assertFalse(line3.has(y))
        self.assertTrue(line3.has(0))
        self.assertFalse(line3.has(1))
        self.assertTrue(line4.has("fun"))

    def test___getattr__(self):
        program = ast.parse("x = 0")
        program = CaitNode(program)
        self.assertTrue(program.ast_name == "Module", "Expected ast_name of 'Module', not{}".format(program.ast_name))

        assign = program.children[0]
        assign_value = assign.value
        self.assertTrue(assign_value.ast_name == "Num" or assign_value.ast_name == "Constant",
                        "__getattribute__ fallthrough failed")

        assign_targets = assign.targets
        self.assertTrue(type(assign_targets) == list, "expected list, got {} instead".format(list))

        assign_target = assign.target
        self.assertTrue(type(assign_target) == CaitNode, "Expected CaitNode, got {} instead".format(assign_target))

        program1 = ast.parse("0 < x < 1")
        program1 = CaitNode(program1)
        binops_node = program1.children[0].value

        binops_names = binops_node.ops_names
        binops_funcs = binops_node.ops
        self.assertTrue(type(binops_funcs) == list, "Expected list, got {} instead".format(type(binops_funcs)))
        self.assertTrue(type(binops_names) == list, "Expected list, got {} instead".format(type(binops_funcs)))
        self.assertTrue(isinstance(binops_funcs[0].astNode, ast.cmpop),
                        "Expected ast.cmpop, got {} instead".format(type(binops_funcs[0])))
        self.assertTrue(type(binops_names[0]) == str, "Expected ast.cmpop")

    @unittest.skip("Not implemented yet")
    def test_numeric_logic_check(self):
        program = CaitNode(ast.parse("if 24 < x < 35:\n"
                                     "    pass"))
        compare = program.linear_tree[2]
        self.assertTrue(compare.numeric_logic_check(1, "24 < x < 35"))
        self.assertFalse(compare.numeric_logic_check(1, "23 < x < 35"))

        program = CaitNode(ast.parse("if '24' < x < '35':\n"
                                     "    pass"))
        compare = program.linear_tree[2]
        self.assertFalse(compare.numeric_logic_check(1, "23 < x < 35"))

        program = CaitNode(ast.parse("20 < x and x < 25 or 15 < x and 15 < x < 32"))
        compare = program.body[0].value
        self.assertTrue(compare.numeric_logic_check(1, "20 < x < 25 or 15 < x < 32"))

        program = CaitNode(ast.parse("12*2 < x*55 + 36 < 55"))
        compare = program.body[0].value
        self.assertTrue(compare.numeric_logic_check(1, "-12 < x*55 < 55"))

        program = CaitNode(ast.parse("temperature >= 32 and temperature <= 50"))
        compare = program.body[0].value
        self.assertTrue(compare.numeric_logic_check(1, "32 <= temp <= 50"))

    def test_get_value(self):
        x_val = 0
        program = ast.parse("x = {x_val}".format(x_val=x_val))
        program = CaitNode(program)
        num_node = program.linear_tree[4]
        self.assertTrue(num_node.value == x_val,
                        ("get_value didn't function as intended "
                         "returned {} instead of {x_val}").format(
                            num_node.value, x_val=x_val))

        x_val = 1.7
        program = ast.parse("x = {x_val}".format(x_val=x_val))
        program = CaitNode(program)
        num_node = program.linear_tree[4]
        self.assertTrue(num_node.value == x_val,
                        ("get_value didn't function as intended "
                         "returned {} instead of {x_val}").format(
                            num_node.value, x_val=x_val))

        x_val = "'fun'"
        x_val_s = 'fun'
        program = ast.parse("x = {x_val}".format(x_val=x_val))
        program = CaitNode(program)
        str_node = program.linear_tree[4]
        self.assertTrue(str_node.value == x_val_s,
                        ("get_value didn't function as intended "
                         "returned {} instead of {x_val}").format(
                            str_node.value, x_val=x_val_s))

        x_val = False
        program = ast.parse("x = {x_val}".format(x_val=x_val))
        program = CaitNode(program)
        num_node = program.linear_tree[4]
        self.assertTrue(num_node.value == x_val,
                        ("get_value didn't function as intended "
                         "returned {} instead of {x_val}").format(
                            num_node.value, x_val=x_val))

        # TODO: Test .value for Name nodes

    def test_nested_expr_matching(self):
        student_code = ("my_sum = 0\n"
                        "for my_item in my_list:\n"
                        "    if True:\n"
                        "        my_sum = my_sum + my_item\n")
        pattern0 = ("for _item_ in _list_:\n"
                    "    __expr__")
        match01 = find_match(pattern0, student_code)
        self.assertTrue(match01)

        __expr__ = match01['__expr__']
        match02 = __expr__.find_match("___ = ___ + _item_")
        self.assertTrue(match02)

        # Case 2
        student_code = ("my_sum = 0\n"
                        "for my_item in my_list:\n"
                        "    if True:\n"
                        "        my_sum = my_sum + my_item2\n")
        pattern0 = ("for _item_ in _list_:\n"
                    "    __expr__")
        match01 = find_match(pattern0, student_code)
        self.assertTrue(match01)

        __expr__ = match01['__expr__']
        match02 = __expr__.find_match("___ = ___ + _item_")
        self.assertFalse(match02)



import ast

from pedal.types.new_types import (AnyType, ImpossibleType,
                                   NumType, BoolType,
                                   TupleType, ListType, StrType,
                                   DictType, SetType, GeneratorType,
                                   FunctionType,
                                   InstanceType, ClassType, LiteralValue,
                                   IntType, FloatType)


def add_tuples(left, right):
    """ Literally just concatenate the types """
    return tuple(left.element_types) + tuple(right.element_types)


def add_element_container_types(left, right):
    """ Use whichever type is not empty """
    if left.is_empty:
        return right.clone()
    else:
        return left.clone()
    #return right.element_type.clone() if left.is_empty else left.element_type.clone()


def NumType_any(*x):
    """ Ignores all parameters to return a NumType """
    return NumType()


def FloatType_any(*x):
    return FloatType()


def IntType_any(*x):
    return IntType()


def StrType_any(*x):
    """ Ignores all parameters to return a StrType """
    return StrType(False)


def BoolType_any(*x):
    """ Ignores all parameters to return a BoolType """
    return BoolType()


def keep_left(left, right):
    """ Returns the left parameter """
    return left


def keep_right(left, right):
    """ Returns the right parameter """
    return right


# Maps the operations to their return types, based on the values.
# The *_any functions don't use the parameters, they return values unconditionally.
VALID_BINOP_TYPES = {
    ast.Add: {NumType: {NumType: NumType_any,
                        IntType: NumType_any,
                        FloatType: FloatType_any},
              IntType: {NumType: NumType_any,
                        IntType: IntType_any,
                        FloatType: FloatType_any},
              FloatType: {NumType: NumType_any,
                          IntType: FloatType_any,
                          FloatType: FloatType_any},
              StrType: {StrType: StrType_any},
              ListType: {ListType: add_element_container_types},
              TupleType: {TupleType: add_tuples}},
    ast.Sub: {NumType: {NumType: NumType_any,
                        IntType: NumType_any,
                        FloatType: FloatType_any},
              IntType: {NumType: NumType_any,
                        IntType: IntType_any,
                        FloatType: FloatType_any},
              FloatType: {NumType: NumType_any,
                          IntType: FloatType_any,
                          FloatType: FloatType_any},
              SetType: {SetType: add_element_container_types}},
    ast.Div: {NumType: {NumType: NumType_any,
                        IntType: NumType_any,
                        FloatType: FloatType_any},
              IntType: {NumType: NumType_any,
                        IntType: FloatType_any,
                        FloatType: FloatType_any},
              FloatType: {NumType: NumType_any,
                          IntType: FloatType_any,
                          FloatType: FloatType_any}},
    ast.FloorDiv: {NumType: {NumType: IntType_any,
                             IntType: IntType_any,
                             FloatType: IntType_any},
                   IntType: {NumType: IntType_any,
                             IntType: IntType_any,
                             FloatType: IntType_any},
                   FloatType: {NumType: IntType_any,
                               IntType: IntType_any,
                               FloatType: IntType_any}},
    ast.Mult: {NumType: {NumType: NumType_any,
                         IntType: NumType_any,
                         FloatType: NumType_any,
                         StrType: StrType_any,
                         ListType: keep_right,
                         TupleType: keep_right},
               FloatType: {NumType: NumType_any,
                           IntType: FloatType_any,
                           FloatType: FloatType_any},
               IntType: {NumType: NumType_any,
                         IntType: IntType_any,
                         FloatType: FloatType_any,
                         StrType: keep_right,
                         ListType: keep_right,
                         TupleType: keep_right},
               StrType: {NumType: keep_left,
                         IntType: keep_left},
               ListType: {NumType: keep_left},
               TupleType: {NumType: keep_left}},
    ast.Pow: {NumType: {NumType: NumType_any,
                        IntType: NumType_any,
                        FloatType: NumType_any},
              IntType: {NumType: NumType_any,
                        IntType: IntType_any,
                        FloatType: FloatType_any},
              FloatType: {NumType: NumType_any,
                          IntType: FloatType_any,
                          FloatType: FloatType_any}},
    # TODO: Should we allow old-fashioned string interpolation?
    # Currently, I vote no because it makes the code harder and is bad form.
    ast.Mod: {NumType: {NumType: NumType_any,
                        IntType: NumType_any,
                        FloatType: NumType_any},
              IntType: {NumType: NumType_any,
                        IntType: IntType_any,
                        FloatType: FloatType_any},
              FloatType: {NumType: NumType_any,
                          IntType: FloatType_any,
                          FloatType: FloatType_any}},
    ast.LShift: {NumType: {NumType: NumType_any,
                           IntType: NumType_any},
                 IntType: {NumType: NumType_any,
                           IntType: IntType_any}},
    ast.RShift: {NumType: {NumType: NumType_any,
                           IntType: NumType_any},
                 IntType: {NumType: NumType_any,
                           IntType: IntType_any}},
    ast.BitOr: {NumType: {NumType: NumType_any,
                           IntType: NumType_any},
                 IntType: {NumType: NumType_any,
                           IntType: IntType_any},
                BoolType: {NumType: NumType_any,
                           BoolType: BoolType_any},
                SetType: {SetType: add_element_container_types}},
    ast.BitXor: {NumType: {NumType: NumType_any,
                           IntType: NumType_any},
                 IntType: {NumType: NumType_any,
                           IntType: IntType_any},
                 BoolType: {NumType: NumType_any,
                            BoolType: BoolType_any},
                 SetType: {SetType: add_element_container_types}},
    ast.BitAnd: {NumType: {NumType: NumType_any,
                           IntType: NumType_any},
                 IntType: {NumType: NumType_any,
                           IntType: IntType_any},
                 BoolType: {NumType: NumType_any,
                            BoolType: BoolType_any},
                 SetType: {SetType: add_element_container_types}}
}

VALID_UNARYOP_TYPES = {
    ast.UAdd: {NumType: NumType, IntType: IntType, FloatType: FloatType},
    ast.USub: {NumType: NumType, IntType: IntType, FloatType: FloatType},
    ast.Invert: {NumType: NumType, IntType: IntType, FloatType: FloatType}
}


def apply_unary_operation(operation, operand):
    if isinstance(operation, ast.Not):
        return BoolType()
    elif isinstance(operand, AnyType):
        return AnyType()
    operand = operand.promote() if isinstance(operand, LiteralValue) else operand
    if type(operation) in VALID_UNARYOP_TYPES:
        op_lookup = VALID_UNARYOP_TYPES[type(operation)]
        if type(operand) in op_lookup:
            return op_lookup[type(operand)]()
    return ImpossibleType()


def apply_binary_operation(operation, left, right):
    if isinstance(left, AnyType):
        return right
    elif isinstance(right, AnyType):
        return left
    left = left.promote() if isinstance(left, LiteralValue) else left
    right = right.promote() if isinstance(right, LiteralValue) else right
    if type(operation) in VALID_BINOP_TYPES:
        op_lookup = VALID_BINOP_TYPES[type(operation)]
        if type(left) in op_lookup:
            op_lookup = op_lookup[type(left)]
            if type(right) in op_lookup:
                op_lookup = op_lookup[type(right)]
                result_type = op_lookup(left, right)
                return result_type
    return ImpossibleType()



from textwrap import dedent
import unittest
from pedal.assertions.static import *
from pedal.utilities.system import IS_AT_LEAST_PYTHON_39
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE


class TestAssertions(ExecutionTestCase):
    def test_prevent_function_usage(self):
        with Execution('sum([1,2,3])') as e:
            self.assertTrue(prevent_function_call('sum'))
        self.assertFeedback(e, "May Not Use Function\n"
                               "You used the function sum on line 1. "
                               "You may not use that function.")

        with Execution('max([1,2,3])') as e:
            self.assertFalse(prevent_function_call('sum'))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_prevent_operator_used(self):
        with Execution('2+3') as e:
            self.assertTrue(prevent_operation('+'))
        self.assertFeedback(e, "May Not Use Operator\n"
                               "You used the operator + on line 1. "
                               "You may not use that operator.")

    def test_prevent_operator_not_used(self):
        with Execution('+3') as e:
            self.assertFalse(prevent_function_call('+'))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_prevent_operator_below_limit(self):
        with Execution('1+3-4') as e:
            self.assertFalse(prevent_function_call('+', at_most=2))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_prevent_operator_above_limit(self):
        with Execution('2+3+4\n1+2') as e:
            self.assertTrue(prevent_operation('+', at_most=2))
        self.assertFeedback(e, "May Not Use Operator\n"
                               "You used the operator + on line 2. "
                               "You may not use that operator more than "
                               "2 times, but you used it 3 times.")

    def test_ensure_operator_above_limit(self):
        with Execution('2+3+4\n1+2') as e:
            self.assertTrue(ensure_operation('+', at_least=5))
        self.assertFeedback(e, "Must Use Operator\n"
                               "You must use the operator + at least "
                               "5 times, but you used it 3 times.")

    def test_prevent_muted(self):
        with Execution('sum([1,2,3])') as e:
            self.assertTrue(prevent_function_call('sum', muted=True))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_prevent_literal_used_int(self):
        with Execution('a = 5\na') as e:
            self.assertTrue(prevent_literal(5))
        self.assertFeedback(e, "May Not Use Literal Value\n"
                               "You used the literal value 5 on line 1. "
                               "You may not use that value.")

    def test_prevent_literal_used_str(self):
        with Execution('print("Hello")') as e:
            self.assertTrue(prevent_literal("Hello"))
        self.assertFeedback(e, "May Not Use Literal Value\n"
                               "You used the literal value 'Hello' on line 1."
                               " You may not use that value.")

    def test_prevent_literal_unused(self):
        with Execution('print("Hello", 5)') as e:
            self.assertFalse(prevent_literal("Fire"))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_prevent_literal_used_negative_int(self):
        with Execution('a = -1+2\na') as e:
            self.assertTrue(prevent_literal(-1))
        self.assertFeedback(e, "May Not Use Literal Value\n"
                               "You used the literal value -1 on line 1. "
                               "You may not use that value.")

    def test_prevent_literal_used_boolean(self):
        with Execution('a = True\na') as e:
            self.assertTrue(prevent_literal(True))
        self.assertFeedback(e, "May Not Use Literal Value\n"
                               "You used the literal value True on line 1. "
                               "You may not use that value.")

    def test_ensure_literal_used_int(self):
        with Execution('a = 6\na') as e:
            self.assertTrue(ensure_literal(5))
        self.assertFeedback(e, "Must Use Literal Value\n"
                               "You must use the literal value 5.")

    def test_ensure_literal_used_str(self):
        with Execution('print("Helo")') as e:
            self.assertTrue(ensure_literal("Hello"))
        self.assertFeedback(e, "Must Use Literal Value\n"
                               "You must use the literal value 'Hello'.")

    def test_ensure_literal_unused(self):
        with Execution('print("Hello", 5)') as e:
            self.assertFalse(ensure_literal("Hello"))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_ensure_literal_used_negative_int(self):
        with Execution('a = -15+2\na') as e:
            self.assertTrue(ensure_literal(-1))
        self.assertFeedback(e, "Must Use Literal Value\n"
                               "You must use the literal value -1.")

    def test_ensure_literal_used_boolean(self):
        with Execution('a = False\na') as e:
            self.assertTrue(ensure_literal(True))
        self.assertFeedback(e, "Must Use Literal Value\n"
                               "You must use the literal value True.")

    @unittest.skipUnless(IS_AT_LEAST_PYTHON_39, "requires python 3.9+")
    def test_ensure_function_types(self):
        # [int]
        with Execution(dedent("""
        def sum_high(numbers: list[int]) -> int:
            total = 0
            for num in numbers:
                if num >= 50:
                    total += num
            return total
        
        sum_high([1,2,3])
        """) )as e:
            self.assertFalse(ensure_function('sum_high', parameters=[[int]], returns='int'))
        self.assertFeedback(e, SUCCESS_MESSAGE)
        # list[int]
        with Execution(dedent("""
        def sum_high(numbers: list[int]) -> int:
            total = 0
            for num in numbers:
                if num >= 50:
                    total += num
            return total

        sum_high([1,2,3])
        """)) as e:
            self.assertFalse(ensure_function('sum_high', parameters=[list[int]], returns='int'))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_prevent_printing_functions(self):
        with Execution(dedent("""def x():\n print('Hello')\nx()""")) as e:
            self.assertTrue(prevent_printing_functions())
        self.assertFeedback(e, """Do Not Print in Function
The function x is printing on line 2. However, that function is not supposed to print.""")

        with Execution(dedent("""def x():\n return 'Hello' \nx()""")) as e:
            self.assertFalse(prevent_printing_functions())
        self.assertFeedback(e, "Complete\nGreat work!")

        with Execution(dedent("""def x():\n return 'Hello' \ndef main():\n  print(x())\nmain()""")) as e:
            self.assertFalse(prevent_printing_functions('main'))
        self.assertFeedback(e, "Complete\nGreat work!")

        with Execution(dedent("""def x():\n return 'Hello' \ndef main():\n  print(x())\nmain()""")) as e:
            self.assertTrue(prevent_printing_functions('x'))
        self.assertFeedback(e, """Do Not Print in Function
The function main is printing on line 4. However, that function is not supposed to print.""")
